'''
======================
Stochastic Translation
======================
'''

import copy
import numpy as np
import random
import logging as log
from arrow import StochasticSystem

from vivarium.core.process import Process
from vivarium.core.experiment import pp
from vivarium_cell.data.amino_acids import amino_acids
from vivarium_cell.data.molecular_weight import molecular_weight
from vivarium_cell.library.polymerize import Elongation, Polymerase, Template, build_stoichiometry, all_products, generate_template

class Ribosome(Polymerase):
    pass

class Transcript(Template):
    pass

def shuffle(l):
    l = [item for item in l]
    np.random.shuffle(l)
    return l

def random_string(alphabet, length):
    string = ''
    for step in range(length):
        string += random.choice(alphabet)
    return string

#: Variable name for unbound ribosomes
UNBOUND_RIBOSOME_KEY = 'Ribosome'

monomer_symbols = []
monomer_ids = []

for symbol, id in amino_acids.items():
    monomer_symbols.append(symbol)
    monomer_ids.append(id)

A = random_string(monomer_symbols, 20)
Z = random_string(monomer_symbols, 60)
B = random_string(monomer_symbols, 30)
Y = random_string(monomer_symbols, 40)

def gather_genes(affinities):
    genes = {}
    for operon, product in affinities.keys():
        if not operon in genes:
            genes[operon] = []
        genes[operon].append(product)
    return genes

def transcripts_to_gene_counts(transcripts, operons):
    counts = {}
    for transcript, genes in operons.items():
        for gene in genes:
            counts[(transcript, gene)] = transcripts.get(transcript, 0)
    return counts

class Translation(Process):

    name = 'translation'
    defaults = {

        'sequences': {
            ('oA', 'eA'): A,
            ('oAZ', 'eA'): A,
            ('oAZ', 'eZ'): Z,
            ('oB', 'eB'): B,
            ('oBY', 'eB'): B,
            ('oBY', 'eY'): Y},

        'templates': {
            ('oA', 'eA'): generate_template(('oA', 'eA'), 20, ['eA']),
            ('oAZ', 'eA'): generate_template(('oAZ', 'eA'), 20, ['eA']),
            ('oAZ', 'eZ'): generate_template(('oAZ', 'eZ'), 60, ['eZ']),
            ('oB', 'eB'): generate_template(('oB', 'eB'), 30, ['eB']),
            ('oBY', 'eB'): generate_template(('oBY', 'eB'), 30, ['eB']),
            ('oBY', 'eY'): generate_template(('oBY', 'eY'), 40, ['eY'])},

        'transcript_affinities': {
            ('oA', 'eA'): 1.0,
            ('oAZ', 'eA'): 2.0,
            ('oAZ', 'eZ'): 5.0,
            ('oB', 'eB'): 1.0,
            ('oBY', 'eB'): 2.0,
            ('oBY', 'eY'): 5.0},

        'elongation_rate': 5.0,
        'polymerase_occlusion': 10,
        'symbol_to_monomer': amino_acids,
        'monomer_ids': monomer_ids,
        'concentration_keys': [],

        'mass_deriver_key': 'mass_deriver',
        'concentrations_deriver_key': 'translation_concentrations',
        'time_step': 1.0,
    }

    def __init__(self, initial_parameters=None):
        '''A stochastic translation model

        .. WARNING::
            Vivarium's knowledge base uses the gene name to name the
            protein. This means that for a gene acrA that codes for
            protein ArcA, you must refer to the gene, transcript, and
            protein each as acrA.

        .. DANGER::
            This documentation will need to be updated to reflect the
            changes in `#185
            <https://github.com/CovertLab/vivarium/pull/185>`_

        :term:`Ports`:

        * **ribosomes**: Expects the ``ribosomes`` variable, whose
          value is a list of the configurations of the ribosomes
          currently active.
        * **molecules**: Expects variables for each of the RNA
          nucleotides.
        * **transcripts**: Expects variables for each transcript to
          translate. Translation will read transcripts from this port.
        * **proteins**: Expects variables for each protein product. The
          produced proteins will be added to this port as counts.
        * **concentrations**: Expects variables for each key in
          ``concentration_keys``. This will be used by a :term:`deriver`
          to convert counts to concentrations.

        Arguments:
            initial_parameters: A dictionary of configuration options.
                Accepts the following keys:

                * **sequences** (:py:class:`dict`): Maps from operon
                  name to the RNA sequence of the operon, as a
                  :py:class:`str`.
                * **templates** (:py:class:`dict`): Maps from the name
                  of an transcript to a :term:`template specification`.
                  The template specification may be generated by
                  :py:func:`cell.library.polymerize.generate_template`
                  like so:

                  >>> from vivarium_cell.library.polymerize import (
                  ...     generate_template)
                  >>> from vivarium.library.pretty import format_dict
                  >>> terminator_index = 5
                  >>> template = generate_template(
                  ...     'oA', terminator_index, ['product1'])
                  >>> print(format_dict(template))
                  {
                      "direction": 1,
                      "id": "oA",
                      "position": 0,
                      "sites": [],
                      "terminators": [
                          {
                              "position": 5,
                              "products": [
                                  "product1"
                              ],
                              "strength": 1.0
                          }
                      ]
                  }


                * **transcript_affinities** (:py:class:`dict`): A map
                  from the name of a transcript to the binding affinity
                  (a :py:class:`float`) of the ribosome for the
                  transcript.
                * **elongation_rate** (:py:class:`float`): The
                  elongation rate of the ribosome.

                  .. todo:: Units of elongation rate

                * **polymerase_occlusion** (:py:class:`int`): The number
                  of base pairs behind the polymerase where another
                  polymerase is occluded and so cannot bind.
                * **symbol_to_monomer** (:py:class:`dict`): Maps from
                  the symbols used to represent monomers in the RNA
                  sequence to the name of the free monomer. This should
                  generally be
                  :py:data:`cell.data.amino_acids.amino_acids`.
                * **monomer_ids** (:py:class:`list`): A list of the
                  names of the free monomers consumed by translation.
                  This can generally be computed as:

                  >>> import pprint
                  >>>
                  >>> from vivarium_cell.data.amino_acids import amino_acids
                  >>> monomer_ids = amino_acids.values()
                  >>> pp = pprint.PrettyPrinter()
                  >>> pp.pprint(list(monomer_ids))
                  ['Alanine',
                   'Arginine',
                   'Asparagine',
                   'Aspartate',
                   'Cysteine',
                   'Glutamate',
                   'Glutamine',
                   'Glycine',
                   'Histidine',
                   'Isoleucine',
                   'Leucine',
                   'Lysine',
                   'Methionine',
                   'Phenylalanine',
                   'Proline',
                   'Serine',
                   'Threonine',
                   'Tryptophan',
                   'Tyrosine',
                   'Valine']

                  Note that we only included the `list()` transformation
                  to make the output prettier. The `dict_values` object
                  returned by `.values()` is sufficiently list-like for
                  use here. Also note that :py:mod:`pprint` just makes
                  the output prettier.
                * **concentration_keys** (:py:class:`list`): A list of
                  variables you want to be able to access as
                  concentrations from the *concentrations* port. The
                  actual conversion is handled by a deriver.

        Example configuring the process (uses
        :py:func:vivarium.library.pretty.format_dict):

        >>> from vivarium.library.pretty import format_dict
        >>> from vivarium_cell.data.amino_acids import amino_acids
        >>> from vivarium_cell.library.polymerize import generate_template
        >>> random.seed(0)  # Needed because process is stochastic
        >>> np.random.seed(0)
        >>> configurations = {
        ...     'sequences': {
        ...         ('oA', 'eA'): 'AWDPT',
        ...         ('oAZ', 'eZ'): 'YVEGELENGGMFISC',
        ...     },
        ...     'templates': {
        ...         ('oA', 'eA'): generate_template(('oA', 'eA'), 5, ['eA']),
        ...         ('oAZ', 'eZ'): generate_template(('oAZ', 'eZ'), 15, ['eA', 'eZ']),
        ...     },
        ...     'transcript_affinities': {
        ...         ('oA', 'eA'): 1.0,
        ...         ('oAZ', 'eZ'): 1.0,
        ...     },
        ...     'elongation_rate': 10.0,
        ...     'polymerase_occlusion': 10,
        ...     'symbol_to_monomer': amino_acids,
        ...     'monomer_ids': amino_acids.values(),
        ...     'concentration_keys': []
        ... }
        >>> # make the translation process, and initialize the states
        >>> translation = Translation(configurations)  # doctest:+ELLIPSIS
        >>> states = {
        ...     'ribosomes': {},
        ...     'molecules': {},
        ...     'proteins': {UNBOUND_RIBOSOME_KEY: 2},
        ...     'transcripts': {
        ...         'oA': 10,
        ...         'oAZ': 10,
        ...     }
        ... }
        >>> states['molecules'].update(
        ...     {
        ...         molecule_id: 100
        ...         for molecule_id in translation.monomer_ids
        ...     }
        ... )
        >>> update = translation.next_update(1, states)
        >>> print(update['ribosomes'])
        {'_add': [{'key': 1, 'state': <class 'vivarium_cell.processes.translation.Ribosome'>: {'id': 1, 'state': 'occluding', 'position': 9, 'template': ('oAZ', 'eZ'), 'template_index': 0, 'terminator': 0}}, {'key': 2, 'state': <class 'vivarium_cell.processes.translation.Ribosome'>: {'id': 2, 'state': 'occluding', 'position': 9, 'template': ('oAZ', 'eZ'), 'template_index': 0, 'terminator': 0}}], '_delete': set()}
        '''

        if not initial_parameters:
            initial_parameters = {}

        self.monomer_symbols = list(amino_acids.keys())
        self.monomer_ids = list(amino_acids.values())

        self.default_parameters = copy.deepcopy(self.defaults)

        templates = self.or_default(initial_parameters, 'templates')

        self.default_parameters['protein_ids'] = all_products({
            key: Template(config)
            for key, config in templates.items()})

        self.default_parameters['transcript_order'] = list(
            initial_parameters.get(
                'transcript_affinities',
                self.default_parameters['transcript_affinities']).keys())
        self.default_parameters['molecule_ids'] = self.monomer_ids

        self.parameters = copy.deepcopy(self.default_parameters)
        self.parameters.update(initial_parameters)

        self.sequences = self.parameters['sequences']
        self.templates = self.parameters['templates']

        self.transcript_affinities = self.parameters['transcript_affinities']
        self.operons = gather_genes(self.transcript_affinities)
        self.operon_order = list(self.operons.keys())
        self.transcript_order = self.parameters['transcript_order']
        self.transcript_count = len(self.transcript_order)

        self.monomer_ids = self.parameters['monomer_ids']
        self.molecule_ids = self.parameters['molecule_ids']
        self.molecule_ids.extend(['ATP', 'ADP'])

        self.protein_ids = self.parameters['protein_ids']
        self.symbol_to_monomer = self.parameters['symbol_to_monomer']
        self.elongation = 0
        self.elongation_rate = self.parameters['elongation_rate']
        self.polymerase_occlusion = self.parameters['polymerase_occlusion']
        self.concentration_keys = self.parameters['concentration_keys']

        self.affinity_vector = np.array([
            self.transcript_affinities[transcript_key]
            for transcript_key in self.transcript_order], dtype=np.float64)

        self.stoichiometry = build_stoichiometry(self.transcript_count)

        self.initiation = StochasticSystem(self.stoichiometry)

        self.ribosome_id = 0

        self.protein_keys = self.concentration_keys + self.protein_ids
        self.all_protein_keys = self.protein_keys + [UNBOUND_RIBOSOME_KEY]

        self.mass_deriver_key = self.or_default(initial_parameters, 'mass_deriver_key')
        self.concentrations_deriver_key = self.or_default(
            initial_parameters, 'concentrations_deriver_key')

        log.info('translation parameters: {}'.format(self.parameters))

        super(Translation, self).__init__(self.parameters)

    def ports_schema(self):

        def add_mass(schema, masses, key):
            if '_properties' not in schema:
                schema['_properties'] = {}
            if key in masses:
                schema['_properties']['mw'] = masses[key]
            return schema

        return {
            'ribosomes': {
                '*': {
                    'id': {
                        '_default': -1,
                        '_updater': 'set'},
                    'domain': {
                        '_default': 0,
                        '_updater': 'set'},
                    'state': {
                        '_default': None,
                        '_updater': 'set',
                        '_emit': True},
                    'position': {
                        '_default': 0,
                        '_updater': 'set',
                        '_emit': True},
                    'template': {
                        '_default': None,
                        '_updater': 'set',
                        '_emit': True},
                    'template_index': {
                        '_default': 0,
                        '_updater': 'set',
                        '_emit': True}}},

            'global': {},

            'molecules': {
                molecule: add_mass({
                    '_emit': True,
                    '_default': 0,
                    '_divider': 'split'}, molecular_weight, molecule)
                for molecule in self.molecule_ids},

            'transcripts': {
                transcript: add_mass({
                    '_default': 0,
                    '_divider': 'split'}, molecular_weight, transcript)
                for transcript in list(self.operons.keys())},

            'proteins': {
                protein: add_mass({
                    '_default': 0,
                    '_divider': 'split',
                    '_emit': True}, molecular_weight, protein)
                for protein in self.all_protein_keys},

            'concentrations': {
                molecule: {
                    '_default': 0.0,
                    '_updater': 'set'}
                for molecule in self.protein_keys}}

    def derivers(self):
        return {
            self.mass_deriver_key: {
                'deriver': 'mass_deriver',
                'port_mapping': {
                    'global': 'global'}},
            self.concentrations_deriver_key: {
                'deriver': 'concentrations_deriver',
                'port_mapping': {
                    'global': 'global',
                    'counts': 'proteins',
                    'concentrations': 'concentrations'},
                'config': {
                    'concentration_keys': self.protein_keys}}}

    def next_update(self, timestep, states):
        molecules = states['molecules']
        transcripts = states['transcripts']
        proteins = states['proteins']

        ribosomes = {
            id: Ribosome(ribosome)
            for id, ribosome in states['ribosomes'].items()}

        original_ribosome_keys = ribosomes.keys()

        gene_counts = np.array(
            list(transcripts_to_gene_counts(transcripts, self.operons).values()),
            dtype=np.int64)

        # Find out how many transcripts are currently blocked by a
        # newly initiated ribosome
        bound_transcripts = np.zeros(self.transcript_count, dtype=np.int64)
        ribosomes_by_transcript = {
            transcript_key: []
            for transcript_key in self.transcript_order}
        for ribosome in ribosomes.values():
            ribosomes_by_transcript[ribosome.template].append(ribosome)
        for index, transcript in enumerate(self.transcript_order):
            bound_transcripts[index] = len([
                ribosome
                for ribosome in ribosomes_by_transcript[transcript]
                if ribosome.is_bound()])

        # Make the state for a gillespie simulation out of total number of each
        # transcript not blocked by a bound ribosome, concatenated with the number
        # of each transcript that is bound by a ribosome.
        # These are the two states for each transcript the simulation
        # will operate on, essentially going back and forth between
        # bound and unbound states.

        original_unbound_ribosomes = proteins[UNBOUND_RIBOSOME_KEY]
        monomer_limits = {
            monomer: molecules[monomer]
            for monomer in self.monomer_ids}
        unbound_ribosomes = original_unbound_ribosomes

        templates = {
            key: Template(template)
            for key, template in self.templates.items()}

        time = 0
        now = 0
        elongation = Elongation(
            self.sequences,
            templates,
            monomer_limits,
            self.symbol_to_monomer,
            self.elongation)

        while time < timestep:
            # build the state vector for the gillespie simulation
            substrate = np.concatenate([
                gene_counts - bound_transcripts,
                bound_transcripts,
                [unbound_ribosomes]])

            # find number of monomers until next terminator
            distance = 1 / self.elongation_rate

            # find interval of time that elongates to the point of the next terminator
            interval = min(distance, timestep - time)

            if interval == distance:
                # perform the elongation until the next event
                terminations, monomer_limits, ribosomes = elongation.step(
                    interval,
                    monomer_limits,
                    ribosomes)
                unbound_ribosomes += terminations
            else:
                elongation.store_partial(interval)
                terminations = 0

            # run simulation for interval of time to next terminator
            result = self.initiation.evolve(
                interval,
                substrate,
                self.affinity_vector)

            # go through each event in the simulation and update the state
            ribosome_bindings = 0
            for now, event in zip(result['time'], result['events']):
                # ribosome has bound the transcript
                transcript_key = self.transcript_order[event]
                bound_transcripts[event] += 1

                self.ribosome_id += 1
                new_ribosome = Ribosome({
                    'id': self.ribosome_id,
                    'template': transcript_key,
                    'position': 0})
                new_ribosome.bind()
                new_ribosome.start_polymerizing()
                ribosomes[new_ribosome.id] = new_ribosome

                ribosome_bindings += 1
                unbound_ribosomes -= 1

            # deal with occluding rnap
            for ribosome in ribosomes.values():
                if ribosome.is_unoccluding(self.polymerase_occlusion):
                    bound_transcripts[ribosome.template_index] -= 1
                    ribosome.unocclude()

            time += interval

        # track how far elongation proceeded to start from next iteration
        self.elongation = elongation.elongation - int(elongation.elongation)

        proteins = {
            UNBOUND_RIBOSOME_KEY: unbound_ribosomes - original_unbound_ribosomes}
        proteins.update(elongation.complete_polymers)

        molecules = {
            key: count * -1
            for key, count in elongation.monomers.items()}

        original = set(original_ribosome_keys)
        current = set(ribosomes.keys())
        bound_ribosomes = current - original
        completed_ribosomes = original - current
        continuing_ribosomes = original - completed_ribosomes

        # ATP hydrolysis cost is 2 per amino acid elongation
        molecules['ATP'] = 0
        molecules['ADP'] = 0
        for count in elongation.monomers.values():
            molecules['ATP'] -= 2 * count
            molecules['ADP'] += 2 * count

        ribosome_updates = {
            id: ribosomes[id]
            for id in continuing_ribosomes}

        add_ribosomes = [
            {'key': bound, 'state': ribosomes[bound]}
            for bound in bound_ribosomes]

        ribosome_updates['_add'] = add_ribosomes
        ribosome_updates['_delete'] = completed_ribosomes

        update = {
            'ribosomes': ribosome_updates,
            'molecules': molecules,
            'proteins': proteins}

        return update


def test_translation():
    parameters = {}
    translation = Translation(parameters)

    states = {
        'ribosomes': {},
        'molecules': {'ATP': 100000},
        'proteins': {UNBOUND_RIBOSOME_KEY: 10},
        'transcripts': {
            'oA': 10,
            'oAZ': 10,
            'oB': 10,
            'oBY': 10}}
    states['molecules'].update({
        molecule_id: 100
        for molecule_id in translation.monomer_ids})

    update = translation.next_update(10.0, states)

    pp(update)
    print('complete!')



if __name__ == '__main__':
    test_translation()




'''
====================
Colony Shape Deriver
====================
'''

from __future__ import absolute_import, division, print_function

import alphashape
import numpy as np
from pytest import approx
from shapely.geometry.collection import GeometryCollection
from shapely.geometry.linestring import LineString
from shapely.geometry.multipolygon import MultiPolygon
from shapely.geometry.point import Point
from shapely.geometry.polygon import Polygon

from vivarium.core.experiment import get_in
from vivarium.core.process import Deriver
from vivarium.library.units import units
from vivarium.core.registry import assert_no_divide


class Variables():
    AREA = 'surface_area'
    MASS = 'mass'
    CIRCUMFERENCE = 'circumference'
    MAJOR_AXIS = 'major_axis'
    MINOR_AXIS = 'minor_axis'



def major_minor_axes(shape):
    '''Calculate the lengths of the major and minor axes of a shape

    We assume that the major and minor axes are the dimensions of the
    minimum bounding rectangle of the shape. Note that this is different
    from using PCA to find the axes, especially for highly asymmetrical
    and concave shapes.

    Arguments:
        shape (shapely.polygon.Polygon): The shape to compute axes for.

    Returns:
        tuple: A tuple with the major axis first and the minor axis
        second.
    '''
    rect = shape.minimum_rotated_rectangle
    points = list(rect.exterior.coords)
    points = [np.array(point) for point in points]
    # Shapely returns polygon coordinates in the order in which they
    # would appear while traversing the boundary, so we know that any 3
    # consecutive points span the major and minor axes
    dimension_1 = abs(np.linalg.norm(points[1] - points[0]))
    dimension_2 = abs(np.linalg.norm(points[2] - points[1]))
    major = max(dimension_1, dimension_2)
    minor = min(dimension_1, dimension_2)
    return major, minor


def gen_agent_colony_map(agents, colony_shapes):
    '''Create a map from agent to the colony to which the agent belongs

    An agent is considered within a colony if its ``location``
    :term:`variable` intersects with the colony's interior or border. If
    an agent intersects with multiple colonies, we choose arbitrarily
    which colony the agent belongs to.

    Points containing any nan coordinate are considered to be outside of
    all shapes.

    .. note:: An agent may be part of no colony at all, in which case it
        will not be included in the returned map.

    Arguments:
        agents (dict): Dictionary of ``agents`` :term:`port` state whose
            keys are agent IDs and whose values are agent state
            dictionaries.
        colony_shapes (list): List of polygons that define the colonies.

    Returns:
        dict: Map from agent ID to index of colony in ``colony_shapes``.
    '''
    agent_colony_map = {}
    for agent_id, agent_state in agents.items():
        loc = agent_state['boundary']['location']
        if np.any(np.isnan(loc)):
            continue
        point = Point(*loc)
        for i, colony_shape in enumerate(colony_shapes):
            if colony_shape.intersects(point):
                agent_colony_map[agent_id] = i
                break
    return agent_colony_map


class ColonyShapeDeriver(Deriver):
    '''Derives colony shape metrics from cell locations
    '''

    name = 'colony_shape_deriver'
    defaults = {
        'alpha': 1.0,
        'bounds': [1, 1],
    }

    def ports_schema(self):
        return {
            'agents': {
                '*': {
                    'boundary': {
                        'location': {
                            '_default': [
                                0.5 * bound for bound in
                                self.parameters['bounds']
                            ],
                        },
                        'mass': {
                            '_default': 1339 * units.fg,
                        },
                    },
                },
            },
            'colony_global': {
                'surface_area': {
                    '_default': [],
                    '_updater': 'set',
                    '_divider': assert_no_divide,
                    '_emit': True,
                },
                'major_axis': {
                    '_default': [],
                    '_updater': 'set',
                    '_divider': assert_no_divide,
                    '_emit': True,
                },
                'minor_axis': {
                    '_default': [],
                    '_updater': 'set',
                    '_divider': assert_no_divide,
                    '_emit': True,
                },
                'mass': {
                    '_default': [],
                    '_updater': 'set',
                    '_divider': assert_no_divide,
                    '_emit': True,
                },
                'circumference': {
                    '_default': [],
                    '_updater': 'set',
                    '_divider': assert_no_divide,
                    '_emit': True,
                }
            },
        }

    def next_update(self, timestep, states):
        agents = states['agents']
        points = [
            agent['boundary']['location']
            for agent in agents.values()
        ]
        points = [tuple(point) for point in points if np.all(~np.isnan(point))]
        alpha_shape = alphashape.alphashape(
            set(points), self.parameters['alpha'])
        if isinstance(alpha_shape, Polygon):
            shapes = [alpha_shape]
        elif isinstance(alpha_shape, (Point, LineString)):
            # We need at least 3 cells to form a colony polygon
            shapes = []
        else:
            assert isinstance(
                alpha_shape, (MultiPolygon, GeometryCollection))
            shapes = list(alpha_shape)

        agent_colony_map = gen_agent_colony_map(agents, shapes)

        # Calculate colony surface areas
        areas = [shape.area for shape in shapes]

        # Calculate colony major and minor axes based on bounding
        # rectangles
        major_axes = []
        minor_axes = []
        for shape in shapes:
            if isinstance(shape, Polygon):
                major, minor = major_minor_axes(shape)
                major_axes.append(major)
                minor_axes.append(minor)
            else:
                major_axes.append(0)
                minor_axes.append(0)

        # Calculate colony circumference
        circumference = [shape.length for shape in shapes]

        # Calculate colony masses and cell surface areas
        mass = [0] * len(shapes)
        cell_area = [0] * len(shapes)
        for agent_id, agent_state in agents.items():
            if agent_id not in agent_colony_map:
                # We ignore agents not in any colony
                continue
            colony_index = agent_colony_map[agent_id]
            agent_mass = get_in(agent_state, ('boundary', 'mass'), 0)
            mass[colony_index] += agent_mass

        return {
            'colony_global': {
                'surface_area': areas,
                'major_axis': major_axes,
                'minor_axis': minor_axes,
                'circumference': circumference,
                'mass': mass,
            }
        }

ColonyShapeDeriver()

class TestDeriveColonyShape():

    def calc_shape_metrics(self, points, alpha=None):
        config = {}
        if alpha is not None:
            config['alpha'] = alpha
        deriver = ColonyShapeDeriver(config)
        states = {
            'agents': {
                str(i): {
                    'boundary': {
                        'location': list(point),
                        'mass': 1.0 * units.fg,
                    },
                }
                for i, point in enumerate(points)
            },
            'colony_global': {
                'surface_area': [],
                'major_axis': [],
                'minor_axis': [],
                'circumference': [],
            }
        }
        # Timestep does not matter
        update = deriver.next_update(-1, states)
        return update['colony_global']

    def flatten(self, lst):
        return [
            elem
            for sublist in lst
            for elem in sublist
        ]

    def test_convex(self):
        #    *
        #   / \
        #  * * *
        #   \ /
        #    *
        points = [
            (1, 2),
            (0, 1), (1, 1), (2, 1),
            (1, 0),
        ]
        metrics = self.calc_shape_metrics(points)
        assert metrics['surface_area'] == [2]
        assert metrics['major_axis'] == approx([np.sqrt(2)])
        assert metrics['minor_axis'] == approx([np.sqrt(2)])
        assert metrics['circumference'] == approx([4 * np.sqrt(2)])
        assert metrics['mass'] == [5 * units.fg]

    def test_concave(self):
        # *-*-*-*-*
        # |       |
        # * * *-*-*
        # |  /
        # * *
        # |  \
        # * * *-*-*
        # |       |
        # *-*-*-*-*
        points = (
            [(i, 4) for i in range(5)]
            + [(i, 3) for i in range(5)]
            + [(i, 2) for i in range(2)]
            + [(i, 1) for i in range(5)]
            + [(i, 0) for i in range(5)]
        )
        metrics = self.calc_shape_metrics(points)
        assert metrics['surface_area'] == [11]
        assert metrics['major_axis'] == [4]
        assert metrics['minor_axis'] == [4]
        assert metrics['circumference'] == approx([18 + 2 * np.sqrt(2)])
        assert metrics['mass'] == [22 * units.fg]

    def test_ignore_outliers_and_nan(self):
        #    *
        #   / \
        #  * * *            *
        #   \ /
        #    *
        points = [
            (1, 2),
            (0, 1), (1, 1), (2, 1), (10, 1),
            (1, 0),
            (np.nan, np.nan),
        ]
        metrics = self.calc_shape_metrics(points)
        assert metrics['surface_area'] == [2]
        assert metrics['major_axis'] == approx([np.sqrt(2)])
        assert metrics['minor_axis'] == approx([np.sqrt(2)])
        assert metrics['circumference'] == approx([4 * np.sqrt(2)])
        assert metrics['mass'] == [5 * units.fg]

    def test_colony_too_diffuse(self):
        #    *
        #
        #  *   *
        #
        #    *
        points = [
            (1, 2),
            (0, 1), (2, 1),
            (1, 0),
        ]
        metrics = self.calc_shape_metrics(points)
        expected_metrics = {
            'surface_area': [],
            'major_axis': [],
            'minor_axis': [],
            'circumference': [],
            'mass': [],
        }
        assert metrics == expected_metrics

    def test_single_cell(self):
        #
        #  *
        #
        points = [
            (1, 1),
        ]
        metrics = self.calc_shape_metrics(points)
        expected_metrics = {
            'surface_area': [],
            'major_axis': [],
            'minor_axis': [],
            'circumference': [],
            'mass': [],
        }
        assert metrics == expected_metrics

    def test_two_cells(self):
        #
        #  *   *
        #
        points = [
            (0, 1), (2, 1),
        ]
        metrics = self.calc_shape_metrics(points)
        expected_metrics = {
            'surface_area': [],
            'major_axis': [],
            'minor_axis': [],
            'circumference': [],
            'mass': [],
        }
        assert metrics == expected_metrics

    def test_no_cells(self):
        points = []
        metrics = self.calc_shape_metrics(points)
        expected_metrics = {
            'surface_area': [],
            'major_axis': [],
            'minor_axis': [],
            'circumference': [],
            'mass': [],
        }
        assert metrics == expected_metrics

    def test_find_multiple_colonies(self):
        #    *          *
        #   / \        / \
        #  * * *      * * *
        #   \ /        \ /
        #    *          *
        points = [
            (1, 2), (11, 2),
            (0, 1), (1, 1), (2, 1), (10, 1), (11, 1), (12, 1),
            (1, 0), (11, 0),
        ]
        metrics = self.calc_shape_metrics(points)
        assert metrics['surface_area'] == [2, 2]
        assert metrics['major_axis'] == approx([np.sqrt(2), np.sqrt(2)])
        assert metrics['minor_axis'] == approx([np.sqrt(2), np.sqrt(2)])
        assert metrics['circumference'] == approx([4 * np.sqrt(2)] * 2)
        assert metrics['mass'] == [5 * units.fg] * 2



"""
========================
Inclusion body composite
========================
"""

import os

from vivarium.library.units import units
from vivarium.library.dict_utils import deep_merge
from vivarium.core.process import Composite
from vivarium.core.composition import (
    simulate_compartment_in_experiment,
    COMPOSITE_OUT_DIR,
)
from vivarium.plots.agents_multigen import plot_agents_multigen
from vivarium.plots.topology import plot_compartment_topology

# processes
from vivarium.processes.meta_division import MetaDivision
from vivarium_cell.processes.inclusion_body import InclusionBody
from vivarium_cell.processes.growth_rate import GrowthRate
from vivarium.processes.divide_condition import DivideCondition
from vivarium_cell.processes.derive_globals import DeriveGlobals



NAME = 'inclusion_body_growth'


class InclusionBodyGrowth(Composite):

    defaults = {
        'inclusion_process': {},
        'growth_rate': {},
        'divide_condition': {
            'threshold': 3000 * units.fg},
        'mass': {},
        'boundary_path': ('boundary',),
        'agents_path': ('..', '..', 'agents',),
        'daughter_path': tuple(),
        'initial_state_config': {
            'inclusion_process': {
                'initial_mass': 10},
            'growth_rate': {
                'initial_mass': 1200}}}

    def __init__(self, config):
        super(InclusionBodyGrowth, self).__init__(config)

    def generate_processes(self, config):
        # division config
        daughter_path = config['daughter_path']
        agent_id = config['agent_id']
        division_config = dict(
            config.get('division', {}),
            daughter_path=daughter_path,
            agent_id=agent_id,
            generator=self)

        return {
            'inclusion_process': InclusionBody(config['inclusion_process']),
            'growth_rate': GrowthRate(config['growth_rate']),
            'globals_deriver': DeriveGlobals({}),
            'divide_condition': DivideCondition(config['divide_condition']),
            'division': MetaDivision(division_config)
        }

    def generate_topology(self, config):
        boundary_path = config['boundary_path']
        agents_path = config['agents_path']
        return {
            'inclusion_process': {
                'front': ('front',),
                'back': ('back',),
                'inclusion_body': ('inclusion_body',),
                'molecules': ('molecules',),
            },
            'growth_rate': {
                'molecules': ('molecules',),
                'global': boundary_path
            },
            'globals_deriver': {
                'global': boundary_path
            },
            'divide_condition': {
                'variable': boundary_path + ('mass',),
                'divide': boundary_path + ('divide',),
            },
            'division': {
                'global': boundary_path,
                'agents': agents_path
            },
        }


    def initial_state(self, config=None):
        if config is None:
            config = {}
        initial_state_config = self.config['initial_state_config']
        config = deep_merge(config, initial_state_config)

        # get the processes
        network = self.generate()
        processes = network['processes']
        topology = network['topology']

        initial_state = {}
        for name, process in processes.items():
            if name in ['inclusion_process', 'growth_rate']:
                process_state = process.initial_state(config.get(name, {}))

                # replace port name with store name
                process_topology = topology[name]
                replace_port_id = {}
                for port_id, state in process_state.items():
                    store_id = process_topology[port_id][0]
                    if port_id is not store_id:
                        replace_port_id[port_id] = store_id
                for port_id, store_id in replace_port_id.items():
                    process_state[store_id] = process_state[port_id]
                    del process_state[port_id]
                initial_state = deep_merge(initial_state, process_state)

        return initial_state


def test_inclusion_body(total_time=1000):
    agent_id = '0'
    parameters = {
        'agent_id': agent_id,
        'inclusion_process': {
            'damage_rate': 1e-4,  # rapid damage
        },
        'growth_rate': {
            'growth_rate': 0.001  # fast growth
        },
    }
    compartment = InclusionBodyGrowth(parameters)

    initial_state = compartment.initial_state()

    # settings for simulation and plot
    settings = {
        'initial_state': {'agents': {agent_id: initial_state}},
        'outer_path': ('agents', agent_id),
        'return_raw_data': True,
        'timestep': 1,
        'total_time': total_time}
    return simulate_compartment_in_experiment(compartment, settings)

def run_compartment(out_dir='out'):
    data = test_inclusion_body(total_time=4000)
    plot_settings = {}
    plot_agents_multigen(data, plot_settings, out_dir)

def plot_inclusion_topology(out_dir='out'):
    # make a topology network plot
    plot_compartment_topology(
        compartment=InclusionBodyGrowth({'agent_id': '1'}),
        settings={},
        out_dir=out_dir)


if __name__ == '__main__':
    out_dir = os.path.join(COMPOSITE_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    plot_inclusion_topology(out_dir=out_dir)
    run_compartment(out_dir=out_dir)



import warnings

import numpy as np

from vivarium.library.units import Quantity, units
from vivarium.core.registry import (
    updater_registry,
    process_registry,
)
from vivarium_cell.library.lattice_utils import (
    get_bin_site,
    get_bin_volume,
    count_to_concentration,
)

# import processes
from vivarium_cell.processes.derive_globals import DeriveGlobals
from vivarium_cell.processes.local_field import LocalField

# register processes
process_registry.register(DeriveGlobals.name, DeriveGlobals)
process_registry.register(LocalField.name, LocalField)

warnings.warn(
    'Vivarium Cell is deprecated. '
    'Please see https://vivarium-collective.github.io for alternatives.'
)



from __future__ import absolute_import, division, print_function

import sys
import os
import argparse

import numpy as np
from scipy import constants
from scipy.ndimage import convolve
import matplotlib.pyplot as plt

from vivarium.core.process import Deriver
from vivarium.core.composition import (
    simulate_process,
    PROCESS_OUT_DIR
)

from vivarium_cell.processes.diffusion_field import plot_fields
from vivarium_cell.plots.multibody_physics import plot_snapshots


NAME = 'static_field'


class StaticField(Deriver):

    name = NAME
    defaults = {
        'molecules': ['glc'],
        'n_bins': [10, 10],
        'bounds': [10, 10],
        'gradient': {},
        'boundary_port': 'boundary',
        'external_key': 'external',
    }

    def __init__(self, initial_parameters=None):
        if initial_parameters is None:
            initial_parameters = {}

        # initial state
        self.molecules = self.or_default(
            initial_parameters, 'molecules')

        # parameters
        self.bounds = self.or_default(
            initial_parameters, 'bounds')
        self.boundary_port = self.or_default(
            initial_parameters, 'boundary_port')
        self.external_key = self.or_default(
            initial_parameters, 'external_key')

        # initialize gradient fields
        self.gradient = self.or_default(
            initial_parameters, 'gradient')

        parameters = {}
        parameters.update(initial_parameters)
        super(StaticField, self).__init__(parameters)

    def ports_schema(self):
        local_concentration_schema = {
            molecule: {
                '_default': 0.0,
                '_updater': 'set'}
            for molecule in self.molecules}

        schema = {}
        glob_schema = {
            '*': {
                self.boundary_port: {
                    'location': {
                        '_default': [0.5 * bound for bound in self.bounds],
                        '_updater': 'set'},
                    self.external_key: local_concentration_schema}}}
        schema['agents'] = glob_schema
        return schema

    def next_update(self, timestep, states):
        agents = states['agents']

        # get each agent's local environment
        local_environments = {}
        for agent_id, state in agents.items():
            location = state[self.boundary_port]['location']
            local_environments[agent_id] = {
                self.boundary_port: {
                    self.external_key: self.get_concentration(location)}}

        return {'agents': local_environments}

    def get_concentration(self, location):
        concentrations = {}
        if self.gradient['type'] == 'gaussian':
            for molecule_id, specs in self.gradient['molecules'].items():
                deviation = specs['deviation']
                dx = location[0] - specs['center'][0] * self.bounds[0]
                dy = location[1] - specs['center'][1] * self.bounds[1]
                distance = np.sqrt(dx ** 2 + dy ** 2)
                concentrations[molecule_id] = gaussian(deviation, (distance/1000))

        elif self.gradient['type'] == 'linear':
            for molecule_id, specs in self.gradient['molecules'].items():
                slope = specs['slope']
                base = specs.get('base', 0.0)
                dx = location[0] - specs['center'][0] * self.bounds[0]
                dy = location[1] - specs['center'][1] * self.bounds[1]
                distance = np.sqrt(dx ** 2 + dy ** 2)
                concentrations[molecule_id] = base + slope * (distance/1000)

        elif self.gradient['type'] == 'exponential':
            for molecule_id, specs in self.gradient['molecules'].items():
                base = specs['base']
                scale = specs.get('scale', 1)
                dx = location[0] - specs['center'][0] * self.bounds[0]
                dy = location[1] - specs['center'][1] * self.bounds[1]
                distance = np.sqrt(dx ** 2 + dy ** 2)
                concentrations[molecule_id] = scale * base ** (distance/1000)
        return concentrations


StaticField()


def get_exponential_config():
    molecule = 'glc'
    center = [0.1, 0.5]
    bounds = [20, 30]
    scale = 1
    base = 0.1
    return {
        'bounds': bounds,
        'molecules': [molecule],
        'gradient': {
            'type': 'exponential',
            'molecules': {
                molecule: {
                    'center': center,
                    'scale': scale,
                    'base': base}}}}

def make_field(config=get_exponential_config()):
    process = StaticField(config)
    molecules = config['molecules']
    bounds = config['bounds']
    bins_per_micron = 1
    n_bins = [bound * bins_per_micron for bound in bounds]
    molecule = molecules[0]  # TODO get all fields

    field = np.zeros((n_bins[0], n_bins[1]), dtype=np.float64)
    for x in range(n_bins[0]):
        for y in range(n_bins[1]):
            location = [x/bins_per_micron, y/bins_per_micron]
            concentration = process.get_concentration(location)
            field[x,y] = concentration[molecule]
    return field

def plot_field(field, out_dir='out'):
    field = np.transpose(field)
    shape = field.shape
    fig = plt.figure()
    im = plt.imshow(field,
                    origin='lower',
                    extent=[0, shape[1], 0, shape[0]],
                    # vmin=vmin,
                    # vmax=vmax,
                    cmap='BuPu')

    fig_path = os.path.join(out_dir, 'field')
    plt.subplots_adjust(wspace=0.7, hspace=0.1)
    plt.savefig(fig_path, bbox_inches='tight')
    plt.close(fig)


if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    field = make_field()
    plot_field(field, out_dir)



"""
===================
Growth rate process
===================
"""

import os

import numpy as np
from scipy import constants

from vivarium.library.units import units
from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process_in_experiment,
    PROCESS_OUT_DIR,
)
from vivarium.plots.simulation_output import plot_simulation_output


NAME = 'growth_rate'
AVOGADRO = constants.N_A


class GrowthRate(Process):
    """The GrowthRate :term:`process class` models exponential cell growth.

    The cell's mass :math:`m_{t + h}` at time :math:`t + h` for
    :term:`timestep` :math:`h` and with growth rate :math:`r` is modeled
    as:

    .. math::

        m_{t + h} = m_t e^{rh}

    Configuration Options:

    * ``growth_rate``: The cell's growth rate :math:`r`. This rate is
      0.0005 by default to approximate an expected 22.5 minutes expected doubling
      time in LB medium: https://bionumbers.hms.harvard.edu/bionumber.aspx?s=n&v=0&id=110058
    """

    name = NAME
    defaults = {
        'growth_rate': 0.0005,
        'mw': {
            'biomass': AVOGADRO * units.fg / units.mol},
        'global_deriver_key': 'globals_deriver',
        'mass_deriver_key': 'mass_deriver',
    }

    def __init__(self, initial_parameters=None):
        super(GrowthRate, self).__init__(initial_parameters)

    def initial_state(self, config=None):
        if config is None:
            config = {}
        initial_mass = config.get('initial_mass', 1339)
        # TODO -- divide up molecules to get initial_mass
        return {
            'molecules': {
                molecule: initial_mass
            } for molecule in self.parameters['mw']
        }

    def ports_schema(self):
        return {
            'molecules': {
                molecule: {
                    '_emit': True,
                    '_default': 1.0,
                    '_divider': 'split',
                    '_properties': {
                        'mw': mw
                    }
                } for molecule, mw in self.parameters['mw'].items()
            },
            'global': {
                'growth_rate': {
                    '_default': self.parameters['growth_rate']
                }
            }
        }

    def derivers(self):
        return {
            self.parameters['mass_deriver_key']: {
                'deriver': 'mass_deriver',
                'port_mapping': {
                    'global': 'global'},
                'config': {}
            }
        }

    def next_update(self, timestep, states):
        molecules = states['molecules']
        growth_rate = states['global']['growth_rate']
        molecules_update = {}
        for molecule, value in molecules.items():
            molecules_update[molecule] = value * np.exp(growth_rate * timestep) - value

        return {
            'molecules': molecules_update
        }


def test_growth_rate():
    growth_rate = GrowthRate({})
    initial_state = growth_rate.initial_state({})
    settings = {
        'total_time': 1350,
        'initial_state': initial_state
    }
    output = simulate_process_in_experiment(growth_rate, settings)

    return output


def run_compartment(out_dir):
    data = test_growth_rate()
    mass = data['global']['mass']
    growth = mass[-1]/mass[0]
    print('growth: {}'.format(growth))

    plot_settings = {}
    plot_simulation_output(data, plot_settings, out_dir)



if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    run_compartment(out_dir)



'''
========================
Stochastic Transcription
========================
'''

import os
import copy
import numpy as np
import logging as log
from arrow import StochasticSystem

from vivarium.library.units import units
from vivarium.library.dict_utils import deep_merge, keys_list
from vivarium.core.experiment import pp, pf
from vivarium.core.process import Process
from vivarium.core.composition import process_in_experiment
from vivarium_cell.states.chromosome import Chromosome, Rnap, Promoter, frequencies, add_merge, toy_chromosome_config
from vivarium_cell.library.polymerize import Elongation, build_stoichiometry, template_products
from vivarium_cell.data.nucleotides import nucleotides

def choose_element(elements):
    if elements:
        choice = np.random.choice(len(elements), 1)
        return list(elements)[int(choice)]

#: Variable name for unbound RNA polymerase
UNBOUND_RNAP_KEY = 'RNA Polymerase'

monomer_ids = list(nucleotides.values())

#: The default configuration parameters for :py:class:`Transcription`
class Transcription(Process):

    name = 'transcription'
    defaults = {
        'promoter_affinities': {},
        'transcription_factors': [],
        'sequence': '',
        'templates': {},
        'genes': {},
        'elongation_rate': 1.0,
        'polymerase_occlusion': 5,
        'symbol_to_monomer': nucleotides,
        'monomer_ids': monomer_ids,
        'concentrations_deriver_key': 'transcription_concentrations',
        'initial_domains': {
            0: {
                'id': 0,
                'lead': 0,
                'lag': 0,
                'children': []}},
        'molecule_ids': monomer_ids,
        'time_step': 1.0,
    }

    def __init__(self, initial_parameters=None):
        '''A stochastic transcription model

        .. WARNING:: Vivarium's knowledge base uses the gene name to
            name the protein. This means that for a gene acrA that
            codes for protein AcrA, you must refer to the gene,
            transcript, and protein each as acrA.

        :term:`Ports`:

        * **chromosome**: The linked :term:`store` should hold a
          representation of the chromosome in the form returned by
          :py:meth:`vivarium_cell.states.chromosome.Chromosome.to_dict`.
        * **molecules**: Expects variables with the names in the
          *molecule_ids* configuration. These are the monomers consumed
          by transcription.
        * **factors**: Expects variables for each transcription factor's
          concentration.
        * **transcripts**: The linked store should store the
          concentrations of the transcripts.
        * **proteins**: The linked store should hold the concentrations
          of the transcription factors and the RNA polymerase.

        Arguments:
            initial_parameters: The following configuration options may
                be provided:

                * **promoter_affinities** (:py:class:`dict`): Maps from
                  binding state tuples to the binding affinity of RNA
                  polymerase and the promoter when the promoter is at
                  that binding state. The binding state of a promoter is
                  which (if any) transcription factors are bound to the
                  promoter. Such a binding state can be represented by a
                  binding state tuple, which is a :py:class:`tuple`
                  whose first element is the name of the promoter. All
                  bound transcription factors are listed as subsequent
                  elements. If no transcription factors are bound, the
                  sole subsequent element is ``None``.

                  .. todo:: What is the significance of the order in the
                      binding state tuple?

                  .. todo:: What are the units of the affinities?

                * **transcription_factors** (:py:class:`list`): A list
                  of all modeled transcription factors.
                * **sequence**: The DNA sequence that includes all the
                  genes whose transcription is being modeled.
                * **templates** (:py:class:`dict`): Maps from the name
                  of an operon to that operon's :term:`template
                  specification`.
                * **genes** (:py:class:`dict`): Maps from operon name to
                  a list of the names of the genes in that operon.
                * **elongation_rate** (:py:class:`float`): The
                  elongation rate of the RNA polymerase.
                * **polymerase_occlusion** (:py:class:`int`): The number
                  of base pairs behind the polymerase where another
                  polymerase is occluded and so cannot bind.
                * **symbol_to_monomer** (:py:class:`dict`): Maps from
                  the symbols used to represent monomers in the RNA
                  sequence to the name of the free monomer. This should
                  generally be
                  :py:data:`cell.data.nucleotides.nucleotides`.
                * **monomer_ids** (:py:class:`list`): A list of the
                  names of the free monomers consumed by transcription.
                  This can generally be computed as:

                  >>> from vivarium_cell.data.nucleotides import nucleotides
                  >>> monomer_ids = nucleotides.values()
                  >>> print(list(monomer_ids))
                  ['ATP', 'GTP', 'UTP', 'CTP']

                  Note that we only included the ``list()``
                  transformation to make the output prettier. The
                  ``dict_values`` object returned by the ``.values()``
                  call is sufficiently list-like for use here.
                * **molecule_ids** (:py:class:`list`): A list of all the
                  molecules needed by the :term:`process`. This will
                  generally be the same as *monomer_ids*.

        Example configuring the process (uses
        :py:func:`vivarium.library.pretty.format_dict`):

        >>> import random
        >>>
        >>> import numpy as np
        >>>
        >>> from vivarium_cell.states.chromosome import (
        ...     toy_chromosome_config,
        ...     Chromosome,
        ... )
        >>> from vivarium_cell.data.nucleotides import nucleotides
        >>> # format_dict lets us print dictionaries prettily
        >>> from vivarium.library.pretty import format_dict
        >>>
        >>> random.seed(0)  # Needed because process is stochastic
        >>> np.random.seed(0)
        >>> # We will use the toy chromosome from toy_chromosome_config
        >>> print(toy_chromosome_config)
        {'sequence': 'ATACGGCACGTGACCGTCAACTTA', 'genes': {'oA': ['eA'], 'oAZ': ['eA', 'eZ'], 'oB': ['eB'], 'oBY': ['eB', 'eY']}, 'promoter_order': ['pA', 'pB'], 'promoters': {'pA': {'id': 'pA', 'position': 3, 'direction': 1, 'sites': [{'position': 0, 'length': 3, 'thresholds': {'tfA': <Quantity(0.3, 'millimolar')>}}], 'terminators': [{'position': 6, 'strength': 0.5, 'products': ['oA']}, {'position': 12, 'strength': 1.0, 'products': ['oAZ']}]}, 'pB': {'id': 'pB', 'position': -3, 'direction': -1, 'sites': [{'position': 0, 'length': 3, 'thresholds': {'tfB': <Quantity(0.5, 'millimolar')>}}], 'terminators': [{'position': -9, 'strength': 0.5, 'products': ['oB']}, {'position': -12, 'strength': 1.0, 'products': ['oBY']}]}}, 'promoter_affinities': {('pA', None): 1.0, ('pA', 'tfA'): 10.0, ('pB', None): 1.0, ('pB', 'tfB'): 10.0}, 'domains': {0: {'id': 0, 'lead': 0, 'lag': 0, 'children': []}}, 'rnaps': {}}
        >>> monomer_ids = list(nucleotides.values())
        >>> configuration = {
        ...     'promoter_affinities': {
        ...         ('pA', None): 1.0,
        ...         ('pA', 'tfA'): 10.0,
        ...         ('pB', None): 1.0,
        ...         ('pB', 'tfB'): 10.0},
        ...     'transcription_factors': ['tfA', 'tfB'],
        ...     'sequence': toy_chromosome_config['sequence'],
        ...     'templates': toy_chromosome_config['promoters'],
        ...     'genes': toy_chromosome_config['genes'],
        ...     'elongation_rate': 10.0,
        ...     'polymerase_occlusion': 5,
        ...     'symbol_to_monomer': nucleotides,
        ...     'monomer_ids': monomer_ids,
        ...     'molecule_ids': monomer_ids,
        ... }
        >>> # At this point we haven't used the toy chromosome yet
        >>> # because it will be specified in the chromosome port.
        >>> # Notice that the parameters are specific to the chromosome.
        >>> transcription_process = Transcription(configuration)
        >>> # Now we need to initialize the simulation stores
        >>> state = {
        ...     'chromosome': toy_chromosome_config,
        ...     'molecules': {
        ...         nucleotide: 10
        ...         for nucleotide in monomer_ids
        ...     },
        ...     'proteins': {UNBOUND_RNAP_KEY: 10},
        ...     'factors': {'tfA': 0.2 * units.mM, 'tfB': 0.7 * units.mM},
        ... }
        >>> update = transcription_process.next_update(1.0, state)
        >>> print(update['chromosome'])
        {'rnaps': {'_add': [{'key': 2, 'state': <class 'vivarium_cell.states.chromosome.Rnap'>: {'id': 2, 'template': 'pA', 'template_index': 0, 'terminator': 1, 'domain': 0, 'state': 'polymerizing', 'position': 7}}, {'key': 3, 'state': <class 'vivarium_cell.states.chromosome.Rnap'>: {'id': 3, 'template': 'pB', 'template_index': 1, 'terminator': 0, 'domain': 0, 'state': 'occluding', 'position': 3}}, {'key': 4, 'state': <class 'vivarium_cell.states.chromosome.Rnap'>: {'id': 4, 'template': 'pA', 'template_index': 0, 'terminator': 0, 'domain': 0, 'state': 'occluding', 'position': 0}}], '_delete': set()}, 'rnap_id': 4, 'domains': {0: <class 'vivarium_cell.states.chromosome.Domain'>: {'id': 0, 'lead': 0, 'lag': 0, 'children': []}}, 'root_domain': 0}
        '''

        if not initial_parameters:
            initial_parameters = {}

        log.debug('inital transcription parameters: {}'.format(initial_parameters))

        super(Transcription, self).__init__(initial_parameters)

        self.derive_defaults('templates', 'promoter_order', keys_list)
        self.derive_defaults('templates', 'transcript_ids', template_products)

        self.sequence = self.parameters['sequence']
        self.templates = self.parameters['templates']
        self.genes = self.parameters['genes']
        empty_chromosome = Chromosome({
            'sequence': self.sequence,
            'promoters': self.templates,
            'genes': self.genes})
        self.sequences = empty_chromosome.sequences()
        self.symbol_to_monomer = self.parameters['symbol_to_monomer']

        log.debug('chromosome sequence: {}'.format(self.sequence))

        self.promoter_affinities = self.parameters['promoter_affinities']
        self.promoter_order = self.parameters['promoter_order']
        self.promoter_count = len(self.promoter_order)

        self.transcription_factors = self.parameters['transcription_factors']
        self.molecule_ids = self.parameters['molecule_ids']
        self.molecule_ids.extend(['ATP', 'ADP'])
        self.monomer_ids = self.parameters['monomer_ids']
        self.transcript_ids = self.parameters['transcript_ids']
        self.elongation = 0
        self.elongation_rate = self.parameters['elongation_rate']
        self.polymerase_occlusion = self.parameters['polymerase_occlusion']

        self.stoichiometry = build_stoichiometry(self.promoter_count)
        self.initiation = StochasticSystem(self.stoichiometry, random_seed=np.random.randint(2**31))

        self.protein_ids = [UNBOUND_RNAP_KEY] + self.transcription_factors

        self.initial_domains = self.parameters['initial_domains']
        self.concentrations_deriver_key = self.parameters['concentrations_deriver_key']

        self.chromosome_ports = ['rnaps', 'rnap_id', 'domains', 'root_domain']

        log.debug('final transcription parameters: {}'.format(self.parameters))

    def build_affinity_vector(self, promoters, factors):
        vector = np.zeros(len(self.promoter_order), dtype=np.float64)
        for index, promoter_key in enumerate(self.promoter_order):
            promoter = promoters[promoter_key]
            binding = promoter.binding_state(factors)
            affinity = self.promoter_affinities.get(binding, 0.0)
            # print('promoter state - {}: {}'.format(binding, affinity))
            vector[index] = affinity
        return vector

    def chromosome_config(self, chromosome_states):
        return dict(
            chromosome_states,
            sequence=self.sequence,
            promoters=self.templates,
            promoter_order=self.promoter_order,
            genes=self.genes)

    def ports_schema(self):
        schema = {}

        schema['chromosome'] = {
            'rnap_id': {
                '_default': 1,
                '_updater': 'set'},
            'root_domain': {
                '_default': 0,
                '_updater': 'set'},
            'domains': {
                '*': {
                    'id': {
                        '_default': 1,
                        '_updater': 'set'},
                    'lead': {
                        '_default': 0,
                        '_updater': 'set'},
                    'lag': {
                        '_default': 0,
                        '_updater': 'set'},
                    'children': {
                        '_default': [],
                        '_updater': 'set'}}},
            'rnaps': {
                '*': {
                    'id': {
                        '_default': -1,
                        '_updater': 'set'},
                    'domain': {
                        '_default': 0,
                        '_updater': 'set'},
                    'state': {
                        '_default': None,
                        '_updater': 'set',
                        '_emit': True},
                    'position': {
                        '_default': 0,
                        '_updater': 'set',
                        '_emit': True},
                    'template': {
                        '_default': None,
                        '_updater': 'set',
                        '_emit': True},
                    'template_index': {
                        '_default': 0,
                        '_updater': 'set',
                        '_emit': True},
                    'terminator': {
                        '_default': 0,
                        '_updater': 'set',
                        '_emit': True}}}}

        initial_domains = {
            id: {
                'id': {
                    '_default': id,
                    '_updater': 'set'},
                'lead': {
                    '_default': 0,
                    '_updater': 'set'},
                'lag': {
                    '_default': 0,
                    '_updater': 'set'},
                'children': {
                    '_default': [],
                    '_updater': 'set'}}
            for id, domain in self.initial_domains.items()}
        schema['chromosome']['domains'].update(initial_domains)

        schema['molecules'] = {
            molecule: {
                '_default': 0,
                '_divider': 'split',
                '_emit': True}
            for molecule in self.molecule_ids}

        schema['factors'] = {
            factor: {
                '_default': 0.0,
                '_divider': 'split'}
            for factor in self.transcription_factors}

        schema['transcripts'] = {
            protein: {
                '_default': 0,
                '_divider': 'split',
                '_emit': True}
            for protein in self.transcript_ids}

        schema['proteins'] = {
            protein: {
                '_default': 0,
                '_divider': 'split',
                '_emit': True}
            for protein in self.protein_ids}

        schema['global'] = {}

        return schema

    def derivers(self):
        return {
            self.concentrations_deriver_key: {
                'deriver': 'concentrations_deriver',
                'port_mapping': {
                    'global': 'global',
                    'counts': 'proteins',
                    'concentrations': 'factors'},
                'config': {
                    'concentration_keys': self.transcription_factors}}}

    def next_update(self, timestep, states):
        chromosome_state = states['chromosome']
        # chromosome_state['rnaps'] = list(chromosome_state['rnaps'].values())
        original_rnap_keys = [
            rnap['id'] for rnap in chromosome_state['rnaps'].values()]
        chromosome = Chromosome(
            self.chromosome_config(chromosome_state))

        molecules = states['molecules']
        proteins = states['proteins']
        factors = states['factors'] # as concentrations

        promoter_rnaps = chromosome.promoter_rnaps()
        promoter_domains = chromosome.promoter_domains()

        # Find out how many promoters are currently blocked by a
        # newly initiated or occluding rnap
        promoter_count = len(chromosome.promoter_order)
        blocked_promoters = np.zeros(promoter_count, dtype=np.int64)
        open_domains = {}
        bound_domains = {}
        for promoter_index, promoter_key in enumerate(chromosome.promoter_order):
            domains = []
            for rnap in promoter_rnaps.get(promoter_key, {}).values():
                if rnap.is_occluding():
                    domains.append(rnap.domain)
                    blocked_promoters[promoter_index] += 1

            bound_domains[promoter_key] = set(domains)
            open_domains[promoter_key] = promoter_domains[promoter_key] - bound_domains[promoter_key]

        blocked_promoters = np.array(blocked_promoters)

        # Make the state for a gillespie simulation out of total number of each
        # promoter by copy number not blocked by initiated rnap,
        # concatenated with the number of each promoter that is bound by rnap.
        # These are the two states for each promoter the simulation
        # will operate on, essentially going back and forth between
        # bound and unbound states.
        copy_numbers = chromosome.promoter_copy_numbers()
        original_unbound_rnaps = proteins[UNBOUND_RNAP_KEY]
        monomer_limits = {
            monomer: molecules[monomer]
            for monomer in self.monomer_ids}
        unbound_rnaps = original_unbound_rnaps

        time = 0
        now = 0
        elongation = Elongation(
            self.sequences,
            chromosome.promoters,
            monomer_limits,
            self.symbol_to_monomer,
            self.elongation)

        initiation_affinity = self.build_affinity_vector(chromosome.promoters, factors)

        while time < timestep:
            # build the state vector for the gillespie simulation
            substrate = np.concatenate([
                copy_numbers - blocked_promoters,
                blocked_promoters,
                [unbound_rnaps]])

            log.debug('transcription substrate: {}'.format(substrate))
            log.debug('blocked promoters: {}'.format(blocked_promoters))

            # find number of monomers until next terminator
            distance = 1 / self.elongation_rate # chromosome.terminator_distance()

            # find interval of time that elongates to the point of the next terminator
            interval = min(distance, timestep - time)

            if interval == distance:
                # perform the elongation until the next event
                terminations, monomer_limits, chromosome.rnaps = elongation.step(
                    interval,
                    monomer_limits,
                    chromosome.rnaps)
                unbound_rnaps += terminations
            else:
                elongation.store_partial(interval)
                terminations = 0

            log.debug('time: {} --- interval: {}'.format(time, interval))
            log.debug('monomer limits: {}'.format(monomer_limits))
            log.debug('terminations: {}'.format(terminations))

            # run simulation for interval of time to next terminator
            result = self.initiation.evolve(
                interval,
                substrate,
                initiation_affinity)

            log.debug('result: {}'.format(result))

            # perform binding
            for now, event in zip(result['time'], result['events']):
                # RNAP has bound the promoter
                promoter_key = chromosome.promoter_order[event]
                promoter = chromosome.promoters[promoter_key]
                domains = open_domains[promoter_key]
                domain = choose_element(domains)

                blocked_promoters[event] += 1
                bound_domains[promoter_key].add(domain)
                open_domains[promoter_key].remove(domain)

                # create a new bound RNAP and add it to the chromosome.
                new_rnap = chromosome.bind_rnap(event, domain)
                new_rnap.start_polymerizing()

                log.debug('newly bound RNAP: {}'.format(new_rnap))

                unbound_rnaps -= 1

            # deal with occluding rnap
            for rnap in chromosome.rnaps.values():
                if rnap.is_unoccluding(self.polymerase_occlusion):
                    log.debug('RNAP unoccluding: {}'.format(rnap))

                    blocked_promoters[rnap.template_index] -= 1
                    bound_domains[rnap.template].remove(rnap.domain)
                    open_domains[rnap.template].add(rnap.domain)
                    rnap.unocclude()
                log.debug('rnap: {}'.format(rnap))

            log.debug('complete: {}'.format(elongation.complete_polymers))

            time += interval

        # track how far elongation proceeded to start from next iteration
        self.elongation = elongation.elongation - int(elongation.elongation)

        proteins = {
            UNBOUND_RNAP_KEY: unbound_rnaps - original_unbound_rnaps}

        molecules = {
            key: count * -1
            for key, count in elongation.monomers.items()}

        # 1 ATP hydrolysis cost per nucleotide elongation
        molecules['ATP'] = 0
        molecules['ADP'] = 0
        for count in elongation.monomers.values():
            molecules['ATP'] -= count
            molecules['ADP'] += count

        chromosome_dict = chromosome.to_dict()
        rnaps = chromosome_dict['rnaps']

        original = set(original_rnap_keys)
        current = set(rnaps.keys())
        bound_rnaps = current - original
        completed_rnaps = original - current
        continuing_rnaps = original - completed_rnaps

        rnap_updates = {
            rnap_id: rnaps[rnap_id]
            for rnap_id in continuing_rnaps}

        add_rnaps = [
            {'key': bound, 'state': rnaps[bound]}
            for bound in bound_rnaps]

        rnap_updates['_add'] = add_rnaps
        rnap_updates['_delete'] = completed_rnaps
        chromosome_dict['rnaps'] = rnap_updates

        update = {
            'chromosome': {
                key: chromosome_dict[key]
                for key in self.chromosome_ports},
            'proteins': proteins,
            'molecules': molecules,
            'transcripts': elongation.complete_polymers}

        log.debug('molecules update: {}'.format(update['molecules']))

        return update


def test_transcription():
    parameters = {
        'sequence': toy_chromosome_config['sequence'],
        'templates': toy_chromosome_config['promoters'],
        'genes': toy_chromosome_config['genes'],
        'promoter_affinities': toy_chromosome_config['promoter_affinities'],
        'transcription_factors': ['tfA', 'tfB'],
        'elongation_rate': 10.0}

    chromosome = Chromosome(toy_chromosome_config)
    transcription = Transcription(parameters)

    initial_molecules = {
                nucleotide: 10
                for nucleotide in transcription.monomer_ids}
    initial_molecules['ATP'] = 100000

    experiment = process_in_experiment(transcription, {
        'initial_state': {
            'chromosome': chromosome.to_dict(),
            'molecules': initial_molecules,
            'proteins': {UNBOUND_RNAP_KEY: 10},
            'factors': {'tfA': 0.2, 'tfB': 0.7}}})

    pp(experiment.state.get_value())
    experiment.update(10.0)
    pp(experiment.state.get_value())

    print('complete!')



if __name__ == '__main__':
    test_transcription()



from __future__ import absolute_import, division, print_function

import os
import copy

from vivarium.library.dict_utils import keys_list
from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process,
    PROCESS_OUT_DIR,
)
from vivarium.plots.simulation_output import plot_simulation_output
from vivarium.library.units import units

from vivarium_cell.data.nucleotides import nucleotides


NAME = 'rna_degradation'

def all_subkeys(d):
    subkeys = set([])
    for ase in d.keys():
        subkeys = subkeys.union(set(d[ase].keys()))
    return list(subkeys)

def kinetics(E, S, kcat, km):
    return kcat * E * S / (S + km)

DEFAULT_TRANSCRIPT_DEGRADATION_KM = 1e-23

TOY_CONFIG = {
    'sequences': {
        'oA': 'GCC',
        'oAZ': 'GCCGUGCAC',
        'oB': 'AGUUGA',
        'oBY': 'AGUUGACGG'
    },
    'catalytic_rates': {
        'endoRNAse': 0.1
    },
    'michaelis_constants': {
        'transcripts': {
            'endoRNAse': {
                'oA': DEFAULT_TRANSCRIPT_DEGRADATION_KM,
                'oAZ': DEFAULT_TRANSCRIPT_DEGRADATION_KM,
                'oB': DEFAULT_TRANSCRIPT_DEGRADATION_KM,
                'oBY': DEFAULT_TRANSCRIPT_DEGRADATION_KM,
            }
        }
    }
}


class RnaDegradation(Process):

    name = NAME
    defaults = {
        'sequences': {},
        'catalytic_rates': {
            'endoRNAse': 0.1},
        'michaelis_constants': {
            'transcripts': {
                'endoRNAse': {}
            }
        },
        'global_deriver_key': 'global_deriver',
        'time_step': 1.0,
    }

    def __init__(self, initial_parameters=None):
        if not initial_parameters:
            initial_parameters = {}

        super(RnaDegradation, self).__init__(initial_parameters)

        self.derive_defaults('sequences', 'transcript_order', keys_list)
        self.derive_defaults('catalytic_rates', 'protein_order', keys_list)

        self.sequences = self.parameters['sequences']
        self.catalytic_rates = self.parameters['catalytic_rates']
        self.michaelis_constants = self.parameters['michaelis_constants']
        self.transcript_order = self.parameters['transcript_order']
        self.protein_order = self.parameters['protein_order']
        self.molecule_order = list(nucleotides.values())
        self.molecule_order.extend(['ATP', 'ADP'])

        self.partial_transcripts = {
            transcript: 0
            for transcript in self.transcript_order}

        self.global_deriver_key = self.or_default(
            initial_parameters, 'global_deriver_key')


    def ports_schema(self):

        ports = [
            'transcripts',
            'proteins',
            'molecules',
            'global']
        schema = {port: {} for port in ports}

        # transcripts
        for state in self.transcript_order:
            schema['transcripts'][state] = {
                '_default': 0,
                '_emit': True}

        # proteins
        for state in self.protein_order:
            schema['proteins'][state] = {
                '_default': 0,
                '_emit': True}

        # molecules
        for state in self.molecule_order:
            schema['molecules'][state] = {
                '_default': 0,
                '_emit': True}

        # global
        schema['global']['mmol_to_counts'] = {
            '_default': 1.0 * units.L / units.mmol}

        return schema

    def derivers(self):
        return {
            self.global_deriver_key: {
                'deriver': 'globals_deriver',
                'port_mapping': {
                    'global': 'global'},
                'config': {
                    'width': 1.11}}}

    def next_update(self, timestep, states):
        transcripts = states['transcripts']
        proteins = states['proteins']
        molecules = states['molecules']
        mmol_to_counts = states['global']['mmol_to_counts']

        delta_transcripts = {
            transcript: 0
            for transcript in self.transcript_order}

        for protein, kcat in self.catalytic_rates.items():
            for transcript, km in self.michaelis_constants['transcripts'][protein].items():
                km *= units.mole / units.fL
                delta_transcripts[transcript] += kinetics(
                    proteins[protein] / mmol_to_counts,
                    transcripts[transcript] / mmol_to_counts,
                    kcat,
                    km)

        degradation_levels = {
            transcript: (
                level * mmol_to_counts * timestep).magnitude + self.partial_transcripts[transcript]
            for transcript, level in delta_transcripts.items()}

        transcript_counts = {
            transcript: -int(level)
            for transcript, level in degradation_levels.items()}

        self.partial_transcripts = {
            transcript: level - int(level)
            for transcript, level in degradation_levels.items()}

        delta_molecules = {
            molecule: 0
            for molecule in self.molecule_order}

        for transcript, count in transcript_counts.items():
            sequence = self.sequences[transcript]
            for base in sequence:
                delta_molecules[nucleotides[base]] -= count
                # ATP hydrolysis cost is 1 per nucleotide degraded
                delta_molecules['ATP'] -= count
                delta_molecules['ADP'] += count

        return {
            'transcripts': transcript_counts,
            'molecules': delta_molecules}


def test_rna_degradation(end_time=10):
    rna_degradation = RnaDegradation(TOY_CONFIG)

    # initial state
    proteins = {
        protein: 10
        for protein in rna_degradation.protein_order}
    molecules = {
        molecule: 10
        for molecule in rna_degradation.molecule_order}
    molecules['ATP'] = 100000
    transcripts = {
        transcript: 10
        for transcript in rna_degradation.transcript_order}

    settings = {
        'total_time': end_time,
        'initial_state': {
            'molecules': molecules,
            'proteins': proteins,
            'transcripts': transcripts}}

    return simulate_process(rna_degradation, settings)


if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    plot_settings = {}
    timeseries = test_rna_degradation(100)
    plot_simulation_output(timeseries, plot_settings, out_dir)



from __future__ import absolute_import, division, print_function

import os

import matplotlib.pyplot as plt

from vivarium.library.units import units
from vivarium_cell.processes.derive_globals import AVOGADRO


def plot_exchanges(timeseries, plot_config, out_dir='out', filename='exchanges'):
    # plot exchanges with the environment

    nAvogadro = AVOGADRO
    external_ts = timeseries['external']
    internal_ts = timeseries['internal']
    global_ts = timeseries['global']

    env_volume = plot_config['environment']['volume']
    legend_on = plot_config.get('legend', True)
    aspect_ratio = plot_config.get('aspect_ratio', 1)

    # pull volume and mass out from internal
    volume = global_ts.pop('volume') * units.fL
    mass = global_ts.pop('mass')

    # conversion factor
    mmol_to_counts = [nAvogadro.to('1/mmol') * vol.to('L') for vol in volume]

    # plot results
    cols = 1
    rows = 3

    width = 5
    height = width / aspect_ratio
    plt.figure(figsize=(width, height))

    # define subplots
    ax1 = plt.subplot(rows, cols, 1)
    ax2 = plt.subplot(rows, cols, 2)
    ax3 = plt.subplot(rows, cols, 3)

    # plot external state
    for mol_id, series in external_ts.items():
        ax1.plot(series, label=mol_id)
    if legend_on:
        ax1.legend(loc='center left', bbox_to_anchor=(1.0, 0.5), ncol=2)
    ax1.title.set_text('environment: {}'.format(env_volume))
    ax1.set_ylabel('external \n (mM)')
    ax1.set_yscale('log')

    # plot internal counts
    for mol_id, counts_series in internal_ts.items():
        # conc_series = [(count / conversion).to('mmol/L').magnitude
        #    for count, conversion in zip(counts_series, mmol_to_counts)]
        ax2.plot(counts_series, label=mol_id)

    if legend_on:
        ax2.legend(loc='center left', bbox_to_anchor=(1.6, 0.5), ncol=3)
    # ax2.title.set_text('internal metabolites')
    ax2.set_ylabel('metabolites \n (counts)')
    ax2.set_yscale('log')

    # plot mass
    ax3.plot(mass, label='mass')
    ax3.set_ylabel('mass \n (fg)')

    # adjust axes
    for axis in [ax1, ax2, ax3]:
        axis.spines['right'].set_visible(False)
        axis.spines['top'].set_visible(False)

    ax1.set_xticklabels([])
    ax2.set_xticklabels([])
    ax3.set_xlabel('time (s)')

    # save figure
    fig_path = os.path.join(out_dir, filename)
    plt.subplots_adjust(wspace=0.3, hspace=0.3)
    plt.savefig(fig_path, bbox_inches='tight')

# energy carriers in BiGG models
BiGG_energy_carriers = [
    'atp_c',
    'gtp_c',
    'nad_c',
    'nadp_c',
    'fad_c',
]

def energy_synthesis_plot(timeseries, settings, out_dir, figname='energy_use'):
    # plot the synthesis of energy carriers in BiGG model output
    energy_reactions = settings.get('reactions', {})
    saved_reactions = timeseries['reactions']
    time_vec = timeseries['time']

    # get each energy carrier's total flux
    carrier_use = {}
    for reaction_id, coeffs in energy_reactions.items():
        reaction_ts = saved_reactions[reaction_id]

        for mol_id, coeff in coeffs.items():

            # save if energy carrier is used
            if coeff < 0:
                added_flux = [-coeff*ts for ts in reaction_ts]
                if mol_id not in carrier_use:
                    carrier_use[mol_id] = added_flux
                else:
                    carrier_use[mol_id] = [
                        sum(x) for x in zip(carrier_use[mol_id], added_flux)]

    # make the figure
    n_cols = 1
    n_rows = 1
    fig = plt.figure(figsize=(n_cols * 6, n_rows * 2))
    grid = plt.GridSpec(n_rows, n_cols)

    # first subplot
    ax = fig.add_subplot(grid[0, 0])
    for mol_id, series in carrier_use.items():
        ax.plot(time_vec, series, label=mol_id)
    ax.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))
    ax.set_title('energy use')
    ax.set_xlabel('time ($s$)')
    ax.set_ylabel('$(mmol*L^{{{}}}*s^{{{}}})$'.format(-1, -1))  # TODO -- use unit schema in figures

    # save figure
    fig_path = os.path.join(out_dir, figname)
    plt.subplots_adjust(wspace=0.3, hspace=0.5)
    plt.savefig(fig_path, bbox_inches='tight')



from __future__ import absolute_import, division, print_function

import os

import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import pyplot as plt
from matplotlib.patches import Patch
from vivarium.library.make_network import save_network

from vivarium_cell.data.amino_acids import amino_acids
from vivarium_cell.data.nucleotides import nucleotides
from vivarium_cell.processes.transcription import UNBOUND_RNAP_KEY
from vivarium_cell.processes.translation import UNBOUND_RIBOSOME_KEY


def plot_timeseries_heatmaps(timeseries, config, out_dir='out'):
    ''' make a timeseries heatmap for each port specified in config['plot_ports'] '''

    name = config.get('name', 'timeseries')
    plot_ports = config.get('plot_ports', {})
    ports = config.get('ports', {})
    time = timeseries['time']

    def relative_to_max(series):
        relative = max(max(series), 1)
        return [
            value / relative
            for value in series]

    # make timeseries heatmaps
    ts_heatmap = {}
    for port_id, order in plot_ports.items():
        port = timeseries[ports[port_id]]
        var_keys = list(order)

        var_series = [
            relative_to_max(port[key])
            for key in var_keys]

        var_keys.reverse()  # reverse to get proper labeling with imshow

        ts_heatmap[port_id] = {
            'keys': var_keys,
            'timeseries': var_series}

    # make figure for each port in plot_ports
    for port_id, heatmap in ts_heatmap.items():
        n_cols = 1
        n_vars = len(heatmap['keys'])

        fig = plt.figure(figsize=(4 * n_cols, 0.6 * n_vars))

        var_keys = heatmap['keys']
        var_series = heatmap['timeseries']
        n_vars = len(var_keys)
        ax = fig.add_subplot(111)

        im = ax.imshow(var_series,
            extent=[time[0], time[-1], 0, n_vars],
            interpolation='nearest',
            aspect='auto',
            cmap='cividis'
            )
        ax.locator_params(axis='y', nbins=n_vars)

        # set y ticks locations and labels
        y_tick_locs = np.asarray([loc+0.5 for loc in range(n_vars)])
        ax.set_yticks(y_tick_locs)
        ax.set_yticklabels(var_keys)
        ax.set_xlabel('time (s)')

        # colorbar
        cbar = fig.colorbar(im)
        cbar.set_label('relative flourescence', rotation=270, labelpad=20)

        # save figure
        figname = name + '_' + port_id
        fig_path = os.path.join(out_dir, figname)
        plt.savefig(fig_path, bbox_inches='tight')


def gene_network_plot(data, out_dir, filename='gene_network'):
    '''
    Make a gene network plot from configuration data

    - data (dict):
        {
        'operons': operons, the "genes" in a chromosome config with {operon: [genes list]}
        'templates': promoters, the "promoters" in a chromosome config with {promoter: {sites: [], thresholds: []}}
        'complexes': complexes, stoichiometry from a complexation process.
        }

    '''
    node_size = 400
    node_distance = 30
    edge_weight = 1
    iterations = 1000

    operon_suffix = '_o'
    tf_suffix = '_tf'
    promoter_suffix = '_p'
    gene_suffix = '_g'
    complex_suffix = '_cxn'

    # plotting parameters
    color_legend = {
        'operon': [x/255 for x in [199,164,53]],
        'gene': [x / 255 for x in [181,99,206]],
        'promoter': [x / 255 for x in [110,196,86]],
        'transcription_factor': [x / 255 for x in [222,85,80]],
        'complex': [x / 255 for x in [153, 204, 255]],
    }

    # get data
    operons = data.get('operons', {})
    templates = data.get('templates', {})
    complexes = data.get('complexes', {})

    # make graph from templates
    G = nx.Graph()

    # initialize node lists for graphing
    operon_nodes = set()
    gene_nodes = set()
    promoter_nodes = set()
    tf_nodes = set()
    complex_nodes = set()

    edges = []

    # add operon --> gene connections
    for operon, genes in operons.items():
        operon_name = operon + operon_suffix
        gene_names = [gene + gene_suffix for gene in genes]

        operon_nodes.add(operon_name)
        gene_nodes.update(gene_names)

        G.add_node(operon_name)
        G.add_nodes_from(gene_names)

        # set node attributes
        operon_attrs = {operon_name: {'type': 'operon'}}
        gene_attrs = {gene: {'type': 'gene'} for gene in gene_names}
        nx.set_node_attributes(G, operon_attrs)
        nx.set_node_attributes(G, gene_attrs)

        # add operon --> gene edge
        for gene_name in gene_names:
            edge = (operon_name, gene_name)
            edges.append(edge)
            G.add_edge(operon_name, gene_name)

    # add transcription factor --> promoter --> operon connections
    for promoter, specs in templates.items():
        promoter_name = promoter + promoter_suffix

        promoter_nodes.add(promoter_name)
        G.add_node(promoter_name)

        # set node attributes
        promoter_attrs = {promoter_name: {'type': 'promoter'}}
        nx.set_node_attributes(G, promoter_attrs)

        # get sites and terminators
        promoter_sites = specs['sites']
        terminators = specs['terminators']

        # add transcription factors
        for site in promoter_sites:
            thresholds = site['thresholds']
            for threshold in thresholds:
                tf_name = threshold + tf_suffix

                tf_nodes.add(tf_name)
                G.add_node(tf_name)

                # set node attributes
                tf_attrs = {tf_name: {'type': 'transcription_factor'}}
                nx.set_node_attributes(G, tf_attrs)

                # connect transcription_factor --> promoter
                edge = (tf_name, promoter_name)
                edges.append(edge)
                G.add_edge(tf_name, promoter_name)

        # add gene products
        for site in terminators:
            products = site['products']
            for product in products:
                operon_name = product + operon_suffix

                operon_nodes.add(operon_name)
                G.add_node(operon_name)

                # set node attributes
                operon_attrs = {operon_name: {'type': 'operon'}}
                nx.set_node_attributes(G, operon_attrs)

                # connect promoter --> operon
                edge = (promoter_name, operon_name)
                edges.append(edge)
                G.add_edge(promoter_name, operon_name)


    ## add gene product --> complex
    # first get the sets of complexes and reactants.
    complex_set = set()
    monomer_set = set()
    for complex, stoichiometry in complexes.items():
        complex_list = [mol_id for mol_id, coeff in stoichiometry.items() if coeff>0]
        reactant_list = [mol_id for mol_id, coeff in stoichiometry.items() if coeff<0]

        complex_set.update(complex_list)
        monomer_set.update(reactant_list)

    # complexes are removed from reactants
    for complex in complex_set:
        if complex in monomer_set:
            monomer_set.remove(complex)

    # add nodes and edges
    for complex, stoichiometry in complexes.items():
        complexes = [mol_id for mol_id, coeff in stoichiometry.items() if coeff>0]
        reactants = {mol_id: coeff for mol_id, coeff in stoichiometry.items() if coeff<0}

        assert len(complexes) == 1, 'too many complexes'
        complex = complexes[0]
        complex_name = complex + complex_suffix

        complex_nodes.add(complex_name)
        G.add_node(complex_name)

        # set node attributes
        complex_attrs = {complex_name: {'type': 'complex'}}
        nx.set_node_attributes(G, complex_attrs)

        # make edge
        for reactant, coeff in reactants.items():

            # is this reactant a monomer or a complex?
            if reactant in monomer_set:
                # TODO -- check that it is actually included in the genes?
                reactant_name = reactant + gene_suffix
            elif reactant in complex_set:
                reactant_name = reactant + complex_suffix

            # connect reactant --> complex
            edge = (reactant_name, complex_name)
            edges.append(edge)
            G.add_edge(reactant_name, complex_name)

    # add edges between proteins/complexes and transcription factors that share the same name
    tf_names = [node.replace(tf_suffix, '') for node in tf_nodes]
    for node in gene_nodes:
        gene_name = node.replace(gene_suffix, '')
        if gene_name in tf_names:
            tf_node = gene_name + tf_suffix
            edge = (node, tf_node)
            edges.append(edge)
            G.add_edge(node, tf_node)

    for node in complex_nodes:
        complex_name = node.replace(complex_suffix, '')
        if complex_name in tf_names:
            tf_node = complex_name + tf_suffix
            edge = (node, tf_node)
            edges.append(edge)
            G.add_edge(node, tf_node)

    # separate out the disconnected graphs
    subgraphs = sorted(nx.connected_components(G), key = len, reverse=True)

    # make node labels by removing suffix
    node_labels = {}
    o_labels = {node: node.replace(operon_suffix,'') for node in operon_nodes}
    g_labels = {node: node.replace(gene_suffix, '') for node in gene_nodes}
    p_labels = {node: node.replace(promoter_suffix, '') for node in promoter_nodes}
    tf_labels = {node: node.replace(tf_suffix, '') for node in tf_nodes}
    cxn_labels = {node: node.replace(complex_suffix, '') for node in complex_nodes}
    node_labels.update(o_labels)
    node_labels.update(g_labels)
    node_labels.update(p_labels)
    node_labels.update(tf_labels)
    node_labels.update(cxn_labels)

    # save network for use in gephi
    gephi_nodes = {}
    gephi_edges = [[node1, node2, edge_weight] for (node1, node2) in edges]
    for node_id in operon_nodes:
        gephi_nodes[node_id] = {
            'label': o_labels[node_id],
            'type': 'operon',
            'size': 1}
    for node_id in gene_nodes:
        gephi_nodes[node_id] = {
            'label': g_labels[node_id],
            'type': 'gene',
            'size': 1}
    for node_id in promoter_nodes:
        gephi_nodes[node_id] = {
            'label': p_labels[node_id],
            'type': 'promoter',
            'size': 1}
    for node_id in tf_nodes:
        gephi_nodes[node_id] = {
            'label': tf_labels[node_id],
            'type': 'transcription factor',
            'size': 1}
    for node_id in complex_nodes:
        gephi_nodes[node_id] = {
            'label': cxn_labels[node_id],
            'type': 'complex',
            'size': 1}
    save_network(gephi_nodes, gephi_edges, out_dir)


    # plot
    n_rows = len(list(subgraphs))
    n_cols = 1
    fig = plt.figure(3, figsize=(6*n_cols, 6*n_rows))
    grid = plt.GridSpec(n_rows, n_cols, wspace=0.2, hspace=0.2)

    for idx, subgraph_nodes in enumerate(subgraphs):
        ax = fig.add_subplot(grid[idx, 0])

        subgraph = G.subgraph(list(subgraph_nodes))
        subgraph_node_labels = {node: node_labels[node] for node in subgraph_nodes}

        # get positions
        dist = node_distance / (len(subgraph)**0.5)
        # pos = nx.spring_layout(subgraph, k=dist, iterations=iterations)
        pos = nx.spring_layout(subgraph,
                               k=dist,
                               scale=20,
                               iterations=iterations)

        color_map = []
        for node in subgraph:
            type = subgraph.nodes[node]['type']
            node_color = color_legend[type]
            color_map.append(node_color)

        nx.draw(subgraph, pos, node_size=node_size, node_color=color_map)

        # edges
        # colors = list(range(1,len(edges)+1))
        nx.draw_networkx_edges(subgraph, pos,
                               # edge_color=colors,
                               width=1.5)

        nx.draw_networkx_labels(subgraph, pos,
                                labels=subgraph_node_labels,
                                font_size=8)

        if idx == 0:
            # make legend
            legend_elements = [Patch(facecolor=color, label=name) for name, color in color_legend.items()]
            plt.legend(handles=legend_elements, bbox_to_anchor=(1.5, 1.0))

    # save figure
    fig_path = os.path.join(out_dir, filename)
    plt.figure(3, figsize=(12, 12))
    plt.axis('off')
    plt.savefig(fig_path, bbox_inches='tight')
    plt.close()


def plot_gene_expression_output(timeseries, config, out_dir='out'):

    name = config.get('name', 'gene_expression')
    ports = config.get('ports', {})
    molecules = timeseries[ports['molecules']]
    transcripts = timeseries[ports['transcripts']]
    proteins = timeseries[ports['proteins']]
    time = timeseries['time']

    # make figure and plot
    n_cols = 1
    n_rows = 5
    plt.figure(figsize=(n_cols * 6, n_rows * 1.5))

    # define subplots
    ax1 = plt.subplot(n_rows, n_cols, 1)
    ax2 = plt.subplot(n_rows, n_cols, 2)
    ax3 = plt.subplot(n_rows, n_cols, 3)
    ax4 = plt.subplot(n_rows, n_cols, 4)
    ax5 = plt.subplot(n_rows, n_cols, 5)

    polymerase_ids = [
        UNBOUND_RNAP_KEY,
        UNBOUND_RIBOSOME_KEY]
    amino_acid_ids = list(amino_acids.values())
    nucleotide_ids = list(nucleotides.values())

    # plot polymerases
    for poly_id in polymerase_ids:
        ax1.plot(time, proteins[poly_id], label=poly_id)
    ax1.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))
    ax1.title.set_text('polymerases')

    # plot nucleotides
    for nuc_id in nucleotide_ids:
        ax2.plot(time, molecules[nuc_id], label=nuc_id)
    ax2.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))
    ax2.title.set_text('nucleotides')

    # plot molecules
    for aa_id in amino_acid_ids:
        ax3.plot(time, molecules[aa_id], label=aa_id)
    ax3.legend(loc='center left', bbox_to_anchor=(2.0, 0.5))
    ax3.title.set_text('amino acids')

    # plot transcripts
    for transcript_id, series in transcripts.items():
        ax4.plot(time, series, label=transcript_id)
    ax4.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))
    ax4.title.set_text('transcripts')

    # plot proteins
    for protein_id in sorted(proteins.keys()):
        if protein_id != UNBOUND_RIBOSOME_KEY:
            ax5.plot(time, proteins[protein_id], label=protein_id)
    ax5.legend(loc='center left', bbox_to_anchor=(1.5, 0.5))
    ax5.title.set_text('proteins')

    # adjust axes
    for axis in [ax1, ax2, ax3, ax4, ax5]:
        axis.spines['right'].set_visible(False)
        axis.spines['top'].set_visible(False)

    ax1.set_xticklabels([])
    ax2.set_xticklabels([])
    ax3.set_xticklabels([])
    ax4.set_xticklabels([])
    ax5.set_xlabel('time (s)', fontsize=12)

    # save figure
    fig_path = os.path.join(out_dir, name)
    plt.subplots_adjust(wspace=0.3, hspace=0.5)
    plt.savefig(fig_path, bbox_inches='tight')


import random

import numpy as np
import logging as log

from vivarium.library.datum import Datum

INFINITY = float('inf')

def flatten(l):
    '''
    Flatten a list by one level:
        [[1, 2, 3], [[4, 5], 6], [7]] --> [1, 2, 3, [4, 5], 6, 7]
    '''

    return [
        item
        for sublist in l
        for item in sublist]

def add_merge(ds):
    '''
    Given a list of dicts, sum the values of each key.
    '''

    result = {}
    for d in ds:
        for key, value in d.items():
            if not key in result:
                result[key] = 0
            result[key] += value
    return result

def kinetics(E, S, kcat, km):
    return kcat * E * S / (S + km)

class Polymerase(Datum):
    defaults = {
        'id': 0,
        'state': None, # other states: ['bound', 'polymerizing', 'complete']
        'position': 0,
        'template': None,
        'template_index': 0,
        'terminator': 0}

    def __init__(self, config):
        super(Polymerase, self).__init__(config)

    def bind(self):
        self.state = 'bound'

    def start_polymerizing(self):
        self.state = 'occluding'

    def complete(self):
        self.state = 'complete'
        # print('completing polymerization: {}'.format(self.to_dict()))

    def is_bound(self):
        return self.state == 'bound'

    def is_polymerizing(self):
        return self.state == 'occluding' or self.state == 'polymerizing'

    def is_complete(self):
        return self.state == 'complete'

    def is_occluding(self):
        return self.state == 'bound' or self.state == 'occluding'

    def is_unoccluding(self, occlusion):
        return self.state == 'occluding' and self.position >= occlusion

    def unocclude(self):
        if self.state == 'occluding':
            self.state = 'polymerizing'

class BindingSite(Datum):
    defaults = {
        'position': 0,
        'length': 0,
        'thresholds': {}} # (factor, threshold)

    def __init__(self, config):
        super(BindingSite, self).__init__(config)

    def state_when(self, levels):
        '''
        Provide the binding state for the given levels of factors. 
        '''

        state = None
        for factor, threshold in self.thresholds.items():
            if levels[factor] >= threshold:
                state = factor
                break
        return state

class Terminator(Datum):
    defaults = {
        'position': 0,
        'strength': 0,
        'products': []}

    def __init__(self, config):
        super(Terminator, self).__init__(config)

    def between(self, before, after):
        return before < self.position < after or after < self.position < before

class Template(Datum):
    schema = {
        'sites': BindingSite,
        'terminators': Terminator}

    defaults = {
        'id': None,
        'position': 0,
        'direction': 1,
        'sites': [],
        'terminators': []}

    def __init__(self, config):
        super(Template, self).__init__(config)

        self.terminator_strength = 0
        for terminator in self.terminators:
            self.terminator_strength += terminator.strength

    def absolute_position(self, relative_position):
        return self.position + (relative_position * self.direction)

    def binding_state(self, levels):
        state = [
            site.state_when(levels)
            for site in self.sites]

        return tuple([self.id] + state)

    def strength_from(self, terminator_index):
        total = 0
        for index in range(terminator_index, len(self.terminators)):
            total += self.terminators[index].strength
        return total

    def next_terminator(self, position):
        for index, terminator in enumerate(self.terminators):
            if terminator.position * self.direction > position * self.direction:
                break
        return index

    def last_terminator(self):
        return self.terminators[-1]

    def terminates_at(self, index=0):
        if len(self.terminators[index:]) > 1:
            choice = random.random() * self.strength_from(index)
            return choice <= self.terminators[index].strength
        else:
            return True

    def choose_terminator(self, index=0):
        if len(self.terminators[index:]) > 1:
            choice = random.random() * self.strength_from(index)
            for terminator in self.terminators[index:]:
                if choice <= terminator.strength:
                    break
                else:
                    choice -= terminator.strength
            return terminator
        else:
            return self.terminators[index]

    def choose_product(self):
        terminator = self.choose_terminator()
        return terminator.products

    def products(self):
        return flatten([
            terminator.products
            for terminator in self.terminators])

def generate_template(id, length, products):
    return {
        'id': id,
        'position': 0,
        'direction': 1,
        'sites': [],
        'terminators': [
            {'position': length,
             'strength': 1.0,
             'products': products}]}

def all_products(templates):
    return list(set([
        product
        for template in templates.values()
        for product in template.products()]))

def template_products(config):
    return all_products({
        key: Template(config)
        for key, config in config.items()})

def polymerize_step(
        sequences,
        polymerases,
        templates,
        symbol_to_monomer,
        monomer_limits):

    complete_polymers = {
        product: 0
        for product in all_products(templates)}

    monomers = {
        monomer: 0
        for monomer in monomer_limits.keys()}
    terminated = 0

    for polymerase in polymerases.values():
        if polymerase.is_polymerizing():
            template = templates[polymerase.template]
            projection = polymerase.position + 1

            try:
                monomer_symbol = sequences[template.id][polymerase.position]
            except IndexError as e:
                log.error('index beyond sequence: polymerase - {} template - {}'.format(
                    polymerase,
                    template))
                monomer_symbol = random.choice(list(symbol_to_monomer.keys()))

            monomer = symbol_to_monomer[monomer_symbol]

            if monomer_limits[monomer] > 0:
                monomer_limits[monomer] -= 1
                monomers[monomer] += 1
                polymerase.position = projection
                absolute_position = template.absolute_position(
                    polymerase.position)

                terminator = template.terminators[polymerase.terminator]
                if terminator.position == absolute_position:
                    if template.terminates_at(polymerase.terminator):
                        polymerase.complete()
                        terminated += 1

                        for product in terminator.products:
                            complete_polymers[product] += 1
                    else:
                        polymerase.terminator += 1

    polymerases = {
        id: polymerase
        for id, polymerase in polymerases.items()
        if not polymerase.is_complete()}

    return monomers, monomer_limits, terminated, complete_polymers, polymerases
    

def polymerize_to(
        sequences,
        polymerases,
        templates,
        additions,
        symbol_to_monomer,
        monomer_limits):

    for step in range(additions):
        monomers, monomer_limits, terminated, complete_polymers, polymerases = polymerize_step(
            sequences, polymerases, templates, symbol_to_monomer, monomer_limits)
    return monomers, monomer_limits, terminated, complete_polymers, polymerases


class Elongation(object):
    def __init__(
            self,
            sequence,
            templates,
            limits,
            symbol_to_monomer,
            elongation=0):
        self.sequence = sequence
        self.templates = templates
        self.time = 0
        self.monomers = {}
        self.symbol_to_monomer = symbol_to_monomer
        self.complete_polymers = {}
        self.previous_elongations = int(elongation)
        self.elongation = elongation
        self.limits = limits

    def step(self, interval, limits, polymerases):
        self.time += interval
        monomers, limits, terminated, complete, polymerases = polymerize_step(
            self.sequence,
            polymerases,
            self.templates,
            self.symbol_to_monomer,
            limits)

        self.monomers = add_merge([self.monomers, monomers])
        self.complete_polymers = add_merge([
            self.complete_polymers, complete])

        return terminated, limits, polymerases

    def store_partial(self, interval):
        self.elongation += interval

    def elongate_to(self, now, rate, limits, polymerases):
        '''
        Track increments of time and accumulate partial elongations, emitting the full
        elongation once a unit is attained.

        Returns number of polymerases that terminated this step, and the updated 
        monomer limits after all elongations.
        '''

        progress = rate * (now - self.time)
        self.elongation += progress
        elongations = int(self.elongation) - self.previous_elongations
        self.time = now
        terminated = 0

        if elongations:
            monomers, limits, terminated, complete, polymerases = polymerize_to(
                self.sequence,
                polymerases,
                self.templates,
                elongations,
                self.symbol_to_monomer,
                limits)
            self.monomers = add_merge([self.monomers, monomers])
            self.complete_polymers = add_merge([
                self.complete_polymers, complete])
            self.previous_elongations = int(self.elongation)

        return terminated, limits, polymerases

    def complete(self):
        return len(self.complete_polymers)


def build_double_stoichiometry(promoter_count):
    '''
    Builds a stoichiometry for the given promoters. There are two states per promoter,
    open and bound, and two reactions per promoter, binding and unbinding. In addition
    there is a single substrate for available RNAP in the final index.

    Here we are assuming
    '''
    stoichiometry = np.zeros((promoter_count * 2, promoter_count * 2 + 1), dtype=np.int64)
    for index in range(promoter_count):
        # forward reaction
        stoichiometry[index][index] = -1
        stoichiometry[index][index + promoter_count] = 1
        stoichiometry[index][-1] = -1 # forward reaction consumes RNAP also

        # reverse reaction
        stoichiometry[index + promoter_count][index] = 1
        stoichiometry[index + promoter_count][index + promoter_count] = -1

    return stoichiometry

def build_double_rates(affinities, advancement):
    return np.concatenate([
        affinities,
        np.repeat(advancement, len(affinities))])

def build_stoichiometry(promoter_count):
    '''
    Builds a stoichiometry for the given promoters. There are two states per promoter,
    open and bound, and two reactions per promoter, binding and unbinding. In addition
    there is a single substrate for available RNAP in the final index.

    Here we are assuming
    '''
    stoichiometry = np.zeros((promoter_count, promoter_count * 2 + 1), dtype=np.int64)
    for index in range(promoter_count):
        # forward reaction
        stoichiometry[index][index] = -1
        stoichiometry[index][index + promoter_count] = 1
        stoichiometry[index][-1] = -1 # forward reaction consumes RNAP also

    return stoichiometry




'''
====================
Convenience Kinetics
====================

Convenience kinetics :cite:`liebermeister_bringing_2006` provides a
generic way to represent rate laws that follow Michaelis-Menten-style
enzyme kinetics. The generalized law can model arbitrary reaction
stoichiometries, catalytic enzymes, activators, and inhibitors.

If you are looking to model a catalyzed process, this may be the
:term:`process class` you need.

Executing this file directly simulates an instance of
:py:class:`ConvenienceKinetics` with the configuration from
:py:func:`get_glc_lct_config`.

------------
Bibliography
------------

.. bibliography:: /references.bib
    :style: plain

'''

import os

import numpy as np

from vivarium.core.process import Process
from vivarium.core.composition import (
    process_in_experiment,
    simulate_experiment,
    flatten_timeseries,
    save_timeseries,
    load_timeseries,
    PROCESS_OUT_DIR,
    assert_timeseries_close,
)
from vivarium.library.dict_utils import tuplify_port_dicts
from vivarium.library.units import units, remove_units
from vivarium.plots.simulation_output import plot_simulation_output

# vivarium-vivarium_cell imports
from vivarium_cell.library.kinetic_rate_laws import KineticFluxModel
from vivarium_cell.data import REFERENCE_DATA_DIR


#: The name of the process, which is used to name the subdirectory under
#: ``out/tests`` that stores the output data from executing this file.
NAME = 'convenience_kinetics'


class ConvenienceKinetics(Process):
    '''Michaelis-Menten-style enzyme kinetics model

     Arguments:
         initial_parameters: Configures the :term:`process` with the
             following configuration options:

             * **reactions** (:py:class:`dict`): Specifies the
               stoichiometry, reversibility, and catalysts of each
               reaction to model. For a non-reversible reaction
               :math:`A + B \\rightleftarrows 2C` catalized by an
               enzyme :math:`E`, we have the following reaction
               specification:

               .. code-block:: python

                 {
                     # reaction1 is a reaction ID
                     'reaction1': {
                         'stoichiometry': {
                             # 1 mol A is consumd per mol reaction
                             ('internal', 'A'): -1,
                             ('internal', 'B'): -1,
                             # 2 mol C are produced per mol reaction
                             ('internal', 'C'): 2,
                         },
                         'is reversible': False,
                         'catalyzed by': [
                             ('internal', 'E'),
                         ],
                     }
                 }

               Note that for simplicity, we assumed all the molecules
               and enzymes were in the ``internal`` port, but this is
               not necessary.
             * **kinetic_parameters** (:py:class:`dict`): Specifies
               the kinetics of the reaction by providing
               :math:`k_{cat}` and :math:`K_M` parameters for each
               enzyme. For example, let's say that for the reaction
               described above, :math:`k{cat} = 1`, :math:`K_A = 2`,
               and :math:`K_B = 3`. Then the reaction kinetics would
               be specified by:

               .. code-block:: python

                 {
                     'reaction1': {
                         ('internal', 'E'): {
                             'kcat_f': 1,  # kcat for forward reaction
                             ('internal', 'A'): 2,
                             ('internal', 'B'): 3,
                         },
                     },
                 }

               If the reaction were reversible, we could have
               specified ``kcat_r`` as the :math:`k_{cat}` of the
               reverse reaction.
             * **initial_state** (:py:class:`dict`): Provides the
               initial quantities of the molecules and enzymes. The
               initial reaction flux must also be specified. For
               example, to start with :math:`[E] = 1.2 mM` and
               :math:`[A] = [B] = [C] = 0 mM`, we would have:

               .. code-block:: python

                 {
                     'internal': {
                         'A': 0.0,
                         'B': 0.0,
                         'C': 0.0,
                         'E': 1.2,
                     },
                 }

               .. note:: Unlike the previous configuration options,
                   the initial state dictionary is not divided up by
                   reaction.

             * **ports** (:py:class:`dict`): Each item in the
               dictionary has a :term:`port` name as its key and a
               list of the :term:`variables` in that port as its
               value. Each port should be specified only once. For
               example, the reaction we have been using as an example
               would have:

               .. code-block:: python

                 {
                     'internal': ['A', 'B', 'C', 'E'],
                 }

     The ports of the process are the ports configured by the
     user, with the following modifications:

     * A ``fluxes`` port is added with variable names equal to
       the IDs of the configured reactions.
     * A ``exchanges`` port is added with the same variables as the
       ``external`` port.
     * A ``global`` port is added with a variable named
       ``mmol_to_counts``, which is set by a :term:`deriver`, and
       ``location``, which is set by the environment.
     * A ``dimensions`` port is added with variables from the
       environment that specify the environment length, width, depth,
       and number of bins.

     Example configuring a process to model the kinetics and reaction
     described above.

     >>> configuration = {
     ...     'reactions': {
     ...         # reaction1 is the reaction ID
     ...         'reaction1': {
     ...             'stoichiometry': {
     ...                 # 1 mol A is consumd per mol reaction
     ...                 ('internal', 'A'): -1,
     ...                 ('internal', 'B'): -1,
     ...                 # 2 mol C are produced per mol reaction
     ...                 ('internal', 'C'): 2,
     ...             },
     ...             'is reversible': False,
     ...             'catalyzed by': [
     ...                 ('internal', 'E'),
     ...             ],
     ...         }
     ...     },
     ...     'kinetic_parameters': {
     ...         'reaction1': {
     ...             ('internal', 'E'): {
     ...                 'kcat_f': 1,  # kcat for forward reaction
     ...                 ('internal', 'A'): 2,
     ...                 ('internal', 'B'): 3,
     ...             },
     ...         },
     ...     },
     ...     'initial_state': {
     ...         'internal': {
     ...             'A': 0.0,
     ...             'B': 0.0,
     ...             'C': 0.0,
     ...             'E': 1.2,
     ...         },
     ...         'fluxes': {
     ...             'reaction1': 0.0,
     ...         }
     ...     },
     ...     'ports': {
     ...         'internal': ['A', 'B', 'C', 'E'],
     ...         'external': [],
     ...     },
     ... }
     >>> kinetic_process = ConvenienceKinetics(configuration)
     '''

    name = NAME
    defaults = {
        'reactions': {},
        'initial_state': {
            'internal': {},
            'external': {}},
        'kinetic_parameters': {},
        'port_ids': [
            'internal',
            'external'
        ],
        'added_port_ids': [
            'fluxes',
            'exchanges',
            'global'
        ],
        'global_deriver_key': 'global_deriver',
        'local_field_key': 'local_field'}

    def __init__(self, parameters=None):
        super(ConvenienceKinetics, self).__init__(parameters)

        self.reactions = self.parameters['reactions']
        kinetic_parameters = self.parameters['kinetic_parameters']
        self.port_ids = self.parameters['port_ids'] + self.parameters['added_port_ids']

        # make the kinetic model
        self.kinetic_rate_laws = KineticFluxModel(self.reactions, kinetic_parameters)

    def initial_state(self, config=None):
        return self.parameters['initial_state']

    def ports_schema(self):

        schema = {port_id: {} for port_id in self.port_ids}
        initial_state = self.initial_state()
        for port, states in initial_state.items():
            for state_id in states:
                schema[port][state_id] = {
                    '_default': initial_state[port][state_id],
                    '_emit': True}

        # exchanges
        # Note: exchanges depends on a port called external
        if 'external' in schema:
            schema['exchanges'] = {
                state_id: {
                    '_default': 0.0,
                }
                for state_id in schema['external'].keys()
            }

        # fluxes
        for state in self.kinetic_rate_laws.reaction_ids:
            schema['fluxes'][state] = {
                '_default': 0.0,
                '_updater': 'set',
            }

        # global
        schema['global'] = {
            'mmol_to_counts': {
                '_default': 0.0 * units.L / units.mmol,
            },
        }

        return schema

    def derivers(self):
        return {
            self.parameters['local_field_key']: {
                'deriver': 'local_field',
                'port_mapping': {
                    'exchanges': 'exchanges',
                },
                'config': {}
            },
            self.parameters['global_deriver_key']: {
                'deriver': 'globals_deriver',
                'port_mapping': {
                    'global': 'global'},
                'config': {
                    'width': 1.0}
            }
        }

    def next_update(self, timestep, states):

        # get mmol_to_counts for converting flux to exchange counts
        mmol_to_counts = states['global']['mmol_to_counts']

        # kinetic rate law requires a flat dict with ('port', 'state') keys.
        flattened_states = remove_units(tuplify_port_dicts(states))

        # get flux, which is in units of mmol / L
        fluxes = self.kinetic_rate_laws.get_fluxes(flattened_states)

        # make the update
        # add fluxes to update
        update = {port: {} for port in self.port_ids}
        update.update({'fluxes': fluxes})

        # update exchanges
        for reaction_id, flux in fluxes.items():
            stoichiometry = self.reactions[reaction_id]['stoichiometry']
            for port_state_id, coeff in stoichiometry.items():
                for port_id in self.port_ids:
                    # separate the state_id and port_id
                    if port_id in port_state_id:
                        state_id = port_state_id[1]
                        state_flux = coeff * flux * timestep

                        if port_id == 'external':
                            # convert exchange fluxes to counts with mmol_to_counts
                            delta = int((state_flux * mmol_to_counts).magnitude)
                            existing_delta = update['exchanges'].get(
                                state_id, {}).get('_value', 0)
                            update['exchanges'][state_id] = existing_delta + delta
                        else:
                            update[port_id][state_id] = (
                                update[port_id].get(state_id, 0)
                                + state_flux
                            )

        # note: external and internal ports update change in mmol.
        return update



# functions
def get_glc_lct_transport():

    transport_reactions = {
        'LCTSt3ipp': {
            'stoichiometry': {
                ('internal', 'h_c'): 1.0,
                ('external', 'h_p'): -1.0,
                ('internal', 'lcts_c'): 1.0,
                ('external', 'lcts_p'): -1.0
            },
            'is reversible': False,
            'catalyzed by': [('internal', 'LacY')]
        },
        'GLCptspp': {
            'stoichiometry': {
                ('internal', 'g6p_c'): 1.0,
                ('external', 'glc__D_e'): -1.0,
                ('internal', 'pep_c'): -1.0,
                ('internal', 'pyr_c'): 1.0,
            },
            'is reversible': False,
            'catalyzed by': [('internal', 'EIIglc')]
        },
        'GLCt2pp': {
            'stoichiometry': {
                ('internal', 'glc__D_c'): 1.0,
                ('external', 'glc__D_p'): -1.0,
                ('internal', 'h_c'): 1.0,
                ('external', 'h_p'): -1.0,
            },
            'is reversible': False,
            'catalyzed by': [('internal', 'GalP')]
        },
    }

    transport_kinetics = {
        # lcts uptake by LacY
        'LCTSt3ipp': {
            ('internal', 'LacY'): {
                ('external', 'h_p'): None,
                ('external', 'lcts_p'): 1e0,
                'kcat_f': 7.8e2,  # 1/s
            }
        },
        # g6p PTS uptake by EIIglc
        'GLCptspp': {
            ('internal', 'EIIglc'): {
                ('external', 'glc__D_e'): 1e0,
                ('internal', 'pep_c'): 1e0,
                'kcat_f': 7.5e4,  # 1/s
            }
        },
        # glc uptake by GalP
        'GLCt2pp': {
            ('internal', 'GalP'): {
                ('external', 'glc__D_p'): 1e0,
                ('external', 'h_p'): None,
                'kcat_f': 1.5e2,  # 1/s
            }
        },
    }

    transport_initial_state = {
        'internal': {
            'EIIglc': 1.8e-3,  # (mmol/L)
            'g6p_c': 0.0,
            'pep_c': 1.8e-1,
            'pyr_c': 0.0,
            'LacY': 0,
            'lcts_p': 0.0,
        },
        'external': {
            'glc__D_e': 10.0,
            'lcts_e': 10.0,
        },
    }

    transport_ports = {
        'internal': [
            'g6p_c', 'pep_c', 'pyr_c', 'EIIglc', 'LacY', 'lcts_p'],
        'external': [
            'glc__D_e', 'lcts_e']
    }

    return {
        'reactions': transport_reactions,
        'kinetic_parameters': transport_kinetics,
        'initial_state': transport_initial_state,
        'ports': transport_ports}


def get_glc_lct_config():
    """
    :py:class:`ConvenienceKinetics` configuration for simplified glucose
    and lactose transport.Glucose uptake simplifies the PTS/GalP system
    to a single uptake kinetic with ``glc__D_e_external`` as the only
    cofactor.

    You can use this configuration with :py:class:`ConvenienceKinetics`
    like this:

    >>> configuration = get_glc_lct_config()
    >>> kinetic_process = ConvenienceKinetics(configuration)
    """
    transport_reactions = {
        'EX_glc__D_e': {
            'stoichiometry': {
                ('internal', 'g6p_c'): 1.0,
                ('external', 'glc__D_e'): -1.0,
                ('internal', 'pep_c'): -1.0,
                ('internal', 'pyr_c'): 1.0,
            },
            'is reversible': False,
            'catalyzed by': [('internal', 'EIIglc')]
        },
        'EX_lcts_e': {
            'stoichiometry': {
                ('external', 'lcts_e'): -1.0,
                ('internal', 'lcts_p'): 1.0,
            },
            'is reversible': False,
            'catalyzed by': [('internal', 'LacY')]
        }
    }
    transport_kinetics = {
        'EX_glc__D_e': {
            ('internal', 'EIIglc'): {
                ('external', 'glc__D_e'): 1e0,  # k_m for external [glc__D_e]
                ('internal', 'pep_c'): None,  # k_m = None makes a reactant non-limiting
                'kcat_f': 1e2,
            }
        },
        'EX_lcts_e': {
            ('internal', 'LacY'): {
                ('external', 'lcts_e'): 1e0,
                'kcat_f': 1e2,
            }
        }
    }

    transport_initial_state = {
        'internal': {
            'EIIglc': 1.0e-3,  # (mmol/L)
            'g6p_c': 0.0,
            'pep_c': 1.0e-1,
            'pyr_c': 0.0,
            'LacY': 0,
            'lcts_p': 0.0,
        },
        'external': {
            'glc__D_e': 10.0,
            'lcts_e': 10.0,
        },
    }

    transport_ports = {
        'internal': [
            'g6p_c', 'pep_c', 'pyr_c', 'EIIglc', 'LacY', 'lcts_p'],
        'external': [
            'glc__D_e', 'lcts_e']
    }

    return {
        'reactions': transport_reactions,
        'kinetic_parameters': transport_kinetics,
        'initial_state': transport_initial_state,
        'ports': transport_ports}


def get_toy_config():
    '''
    Returns
        A configuration for :py:class:`ConvenienceKinetics` that models
        a toy reaction for illustration purposes.
    '''
    toy_reactions = {
        'reaction1': {
            'stoichiometry': {
                ('internal', 'A'): 1,
                ('external', 'B'): -1},
            'is reversible': False,
            'catalyzed by': [('internal', 'enzyme1')]
        }
    }

    toy_kinetics = {
        'reaction1': {
            ('internal', 'enzyme1'): {
                ('external', 'B'): 0.2,
                'kcat_f': 5e1,
            }
        }
    }

    toy_ports = {
        'internal': ['A', 'enzyme1'],
        'external': ['B']
    }

    toy_initial_state = {
        'internal': {
            'A': 1.0,
            'enzyme1': 1e-1,
        },
        'external': {
            'B': 10.0,
        },
    }

    return {
        'reactions': toy_reactions,
        'kinetic_parameters': toy_kinetics,
        'initial_state': toy_initial_state,
        'ports': toy_ports}


def test_convenience_kinetics(end_time=2520):
    config = get_glc_lct_config()
    kinetic_process = ConvenienceKinetics(config)

    initial_state = kinetic_process.initial_state()
    initial_state['external'] = {
            'glc__D_e': 1.0,
            'lcts_e': 1.0}
    settings = {
        'environment': {
            'volume': 1e-14 * units.L,
            'concentrations': initial_state['external'],
        },
        'timestep': 1,
        'total_time': end_time}

    experiment = process_in_experiment(
        process=kinetic_process,
        settings=settings,
        initial_state=initial_state)

    return simulate_experiment(experiment, settings)


def test_convenience_kinetics_correlated_to_reference():
    timeseries = test_convenience_kinetics()
    flattened = flatten_timeseries(timeseries)
    reference_timeseries = load_timeseries(
        os.path.join(REFERENCE_DATA_DIR, NAME + '.csv'))
    assert_timeseries_close(
        flattened, reference_timeseries,
        tolerances={
            'internal_pyr_c': 9,
            'internal_pep_c': 9,
            'internal_g6p_c': 9,
        }
    )


if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    timeseries = test_convenience_kinetics()

    plot_settings = {}
    plot_simulation_output(timeseries, plot_settings, out_dir)
    save_timeseries(timeseries, out_dir)



from vivarium.library.datum import Datum

class Protein(Datum):
    defaults = {
        'id': '',
        'sequence': ''}

    def __init__(self, config):
        super(Protein, self).__init__(config)

GFP = Protein({
    'id': 'GFP',
    'sequence': 'MSKGEELFTGVVPILVELDGDVNGHKFSVSGEGEGDATYGKLTLKFICTTGKLPVPWPTLVTTFSYGVQCFSRYPDHMKQHDFFKSAMPEGYVQERTIFFKDDGNYKTRAEVKFEGDTLVNRIELKGIDFKEDGNILGHKLEYNYNSHNVYIMADKQKNGIKVNFKIRHNIEDGSVQLADHYQQNTPIGDGPVLLPDNHYLSTQSALSKDPNEKRDHMVLLEFVTAAGITHGMDELYK'})




"""
==========================
Inclusion body experiments
==========================
"""

from vivarium.core.control import Control
from vivarium.core.composition import (
    compose_experiment,
    GENERATORS_KEY,
    EXPERIMENT_OUT_DIR,
)

# composites
from vivarium_cell.composites.lattice import (
    Lattice,
    make_lattice_config,
)
from vivarium_cell.composites.inclusion_body_growth import InclusionBodyGrowth

# plots
from vivarium.plots.agents_multigen import plot_agents_multigen
from vivarium_cell.plots.multibody_physics import (
    plot_snapshots,
    plot_tags
)


# def get_experiment_config_1():
#     lattice_config = make_lattice_config(
#         jitter_force=1e-4,
#         bounds=[30, 30],
#         n_bins=[10, 10])
#     return {
#         'lattice_config': lattice_config
#     }

lattice_config = make_lattice_config(
    jitter_force=1e-4,
    bounds=[30, 30],
    n_bins=[10, 10])



def run_experiment(config={}):
    agent_id = '1'
    time_total = 12000

    inclusion_config = {
        'agent_id': agent_id,
        'damage_rate': 5e-5,
    }

    # initial state
    compartment = InclusionBodyGrowth(inclusion_config)
    compartment_state = compartment.initial_state({
        'front': {
            'aggregate': 200},
        'back': {
            'aggregate': 10}
        })
    initial_state = {
        'agents': {
            agent_id: compartment_state}}

    # declare the hierarchy
    hierarchy = {
        GENERATORS_KEY: {
            'type': Lattice,
            'config': lattice_config},
        'agents': {
            agent_id: {
                GENERATORS_KEY: {
                    'type': InclusionBodyGrowth,
                    'config': inclusion_config}}}}

    # configure experiment
    experiment = compose_experiment(
        hierarchy=hierarchy,
        initial_state=initial_state)

    # run simulation
    experiment.update(time_total)
    data = experiment.emitter.get_data()
    experiment.end()

    return data


def inclusion_plots_suite(data=None, out_dir=EXPERIMENT_OUT_DIR):
    n_snapshots = 8
    tagged_molecules = [
        ('inclusion_body',),
        ('front', 'aggregate',),
        ('back', 'aggregate',),
    ]

    # multigen plot
    plot_settings = {}
    plot_agents_multigen(data, plot_settings, out_dir)

    # extract data for snapshots
    multibody_config = lattice_config['multibody']
    agents = {time: time_data['agents'] for time, time_data in data.items()}

    # snapshots plot
    plot_data = {
        'agents': agents,
        'config': multibody_config,
    }
    plot_config = {
        'n_snapshots': n_snapshots,
        'out_dir': out_dir}
    plot_snapshots(plot_data, plot_config)

    # tags plot
    plot_config = {
        'tagged_molecules': tagged_molecules,
        'n_snapshots': n_snapshots,
        'convert_to_concs': False,
        'out_dir': out_dir}
    plot_tags(plot_data, plot_config)



# libraries for control
experiments_library = {
    '1': {
        'name': 'inclusion_lattice',
        'experiment': run_experiment},
}
plots_library = {
    '1': inclusion_plots_suite
}
workflow_library = {
    '1': {
        'name': 'inclusion_body_experiment',
        'experiment': '1',
        'plots': ['1'],
    }
}

if __name__ == '__main__':
    Control(
        experiments=experiments_library,
        plots=plots_library,
        workflows=workflow_library,
        )



'''
=========================
Colony Metrics Experiment
=========================
'''

from __future__ import absolute_import, division, print_function

import os
import random

import numpy as np
import pytest

from vivarium.core.composition import (
    EXPERIMENT_OUT_DIR,
    make_agents,
    assert_timeseries_close,
    load_timeseries,
    simulate_experiment,
)
from vivarium.plots.agents_multigen import plot_agents_multigen
from vivarium.core.emitter import (
    path_timeseries_from_data,
    path_timeseries_from_embedded_timeseries,
    timeseries_from_data,
)
from vivarium.core.experiment import Experiment

from vivarium_cell.composites.lattice import (
    Lattice,
    make_lattice_config,
)
from vivarium_cell.experiments.lattice_experiment import agents_library
from vivarium_cell.plots.multibody_physics import plot_snapshots
from vivarium_cell.plots.colonies import plot_colony_metrics
from vivarium.library.timeseries import (
    process_path_timeseries_for_csv,
    save_flat_timeseries,
)
from vivarium_cell.processes.multibody_physics import single_agent_config
from vivarium_cell.data import REFERENCE_DATA_DIR

NAME = 'colony_metrics'
OUT_DIR = os.path.join(EXPERIMENT_OUT_DIR, NAME)
DEFAULT_BOUNDS = [40, 40]
DEFAULT_EMIT_STEP = 30


def colony_metrics_experiment(config):
    '''Run an experiment to calculate colony metrics

    Arguments:
        config (dict): A dictionary of configuration options which can
            contain the following:

            * **n_agents** (:py:class:`int`): Number of agents to create
              initially
            * **emitter** (:py:class:`dict`): Emitter configuration
              dictionary. This gets passed directly to the
              :py:class:`vivarium.core.experiment.Experiment`
              constructor. Defaults to a timeseries emitter.
            * **environment** (:py:class:`dict`): Configuration to pass
              to the :py:class:`vivarium.composites.lattice.Lattice`
              constructor. Defaults to ``{}``.
            * **agent** (:py:class:`dict`): Dictionary with the
              following keys-value pairs:

                * **compartment**
                  (:py:class:`vivarium.core.experiment.Compartment`): A
                  compartment class to use for each agent.
                * **config** (:py:class:`dict`): Configuration to pass
                  the agent compartment constructor.

            * **locations** (:py:class:`list`): List of sublists. Each
              sublist should have 2 elements, the coordinates at which
              to place an agent. Coordinates range from 0 to 1 and
              represent fractions of environment bounds. Any agents for
              which no location is specified will be placed randomly per
              the default behavior of
              :py:func:`vivarium.processes.multibody.single_agent_config`.

    Returns:
        vivarium.core.experiment.Experiment: An initialized experiment
        object.
    '''
    # configure the experiment
    n_agents = config.get('n_agents')
    emitter = config.get('emitter', {'type': 'timeseries'})

    # make lattice environment
    environment = Lattice(config.get('environment', {}))
    network = environment.generate()
    processes = network['processes']
    topology = network['topology']

    # add the agents
    agent_ids = [str(agent_id) for agent_id in range(n_agents)]
    agent_config = config['agent']
    agent_compartment = agent_config['type']
    compartment_config = agent_config['config']
    agent = agent_compartment(compartment_config)
    agents = make_agents(agent_ids, agent, {})
    processes['agents'] = agents['processes']
    topology['agents'] = agents['topology']

    # initial agent state
    locations = config.get('locations')
    if locations is None:
        locations = [[0.5, 0.5]]
    agent_config_settings = [
        {
            'bounds': environment.config['multibody']['bounds'],
            'location': random.choice(locations) if len(locations) <= index else locations[index]
        }
        for index, agent_id in enumerate(agent_ids)
    ]

    initial_state = {
        'agents': {
            agent_id: single_agent_config(agent_config_setting)
            for agent_id, agent_config_setting
            in zip(agent_ids, agent_config_settings)
        }
    }
    initial_state.update(config.get('initial_state', {}))

    return Experiment({
        'processes': processes,
        'topology': topology,
        'emitter': emitter,
        'initial_state': initial_state,
    })


def get_lattice_with_metrics_config():
    config = {
        'environment': make_lattice_config(
            bounds=DEFAULT_BOUNDS,
        )
    }
    colony_metrics_config = {
        'colony_shape_deriver': {
            'alpha': 1 / 5,
        },
    }
    config['environment'].update(colony_metrics_config)
    return config


def run_experiment(
    runtime=2400, n_agents=2, start_locations=None, growth_rate=0.000275
):
    '''Run a Colony Metrics Experiment

    Arguments:
        runtime (int): Experiment duration
        n_agents (int): Number of agents to create at the start
        start_locations (list): List of initial agent coordinates. If
            you do not provide enough locations, the remaining locations
            will be random. Coordinates range from 0 to 1 and represent
            fractions of environment bounds.

    Returns:
        Simulation data as :term:`raw data`.
    '''
    agent_config = agents_library['growth_division_minimal']
    agent_config['config']['growth_rate_noise'] = 0
    agent_config['growth_rate'] = growth_rate

    experiment_config = get_lattice_with_metrics_config()
    experiment_config.update({
        'n_agents': n_agents,
        'agent': agent_config,
        'locations': start_locations,
    })
    experiment = colony_metrics_experiment(experiment_config)

    # simulate
    settings = {
        'emit_step': DEFAULT_EMIT_STEP,
        'total_time': runtime,
        'return_raw_data': True,
    }
    return simulate_experiment(experiment, settings), experiment_config


@pytest.mark.slow
def test_experiment(seed=1):
    random.seed(seed)
    np.random.seed(seed)
    if not os.path.exists(OUT_DIR):
        os.makedirs(OUT_DIR)
    data, _ = run_experiment(
        start_locations=[[0.3, 0.3], [0.5, 0.5]],
        growth_rate=0.001,
    )
    path_ts = path_timeseries_from_data(data)
    filtered = {
        path: timeseries
        for path, timeseries in path_ts.items()
        # Angles are computed randomly by multibody physics
        if path[-1] != 'angle'
    }
    processed_for_csv = process_path_timeseries_for_csv(filtered)
    save_flat_timeseries(
        processed_for_csv,
        OUT_DIR,
        'test_output.csv'
    )
    test_output = load_timeseries(os.path.join(OUT_DIR, 'test_output.csv'))
    expected = load_timeseries(
        os.path.join(REFERENCE_DATA_DIR, NAME + '.csv'))
    assert_timeseries_close(
        test_output, expected,
        default_tolerance=(1 - 1e-5),
    )


def main():
    if not os.path.exists(OUT_DIR):
        os.makedirs(OUT_DIR)

    data, experiment_config = run_experiment(
        start_locations=[[0.3, 0.3], [0.5, 0.5]],
    )

    # extract data
    multibody_config = experiment_config['environment']['multibody']
    agents = {time: time_data['agents'] for time, time_data in data.items()}
    fields = {time: time_data['fields'] for time, time_data in data.items()}

    # agents plot
    agents_settings = {
        'agents_key': 'agents'
    }
    plot_agents_multigen(data, agents_settings, OUT_DIR, 'agents')

    # snapshot plot
    snapshot_data = {
        'agents': agents,
        'fields': fields,
        'config': multibody_config,
    }
    snapshot_config = {
        'out_dir': OUT_DIR,
        'filename': 'agents_snapshots',
    }
    plot_snapshots(snapshot_data, snapshot_config)

    # Colony Metrics Plot
    embedded_ts = timeseries_from_data(data)
    colony_metrics_ts = embedded_ts['colony_global']
    colony_metrics_ts['time'] = embedded_ts['time']
    path_ts = path_timeseries_from_embedded_timeseries(
        colony_metrics_ts)
    fig = plot_colony_metrics(path_ts)
    fig.savefig(os.path.join(OUT_DIR, 'colonies'))


if __name__ == '__main__':
    main()



"""
======================
Inclusion body process
======================
"""

import os
import random

from scipy import constants

from vivarium.library.units import units
from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process_in_experiment,
    PROCESS_OUT_DIR,
)
from vivarium.plots.simulation_output import plot_simulation_output


NAME = 'inclusion_body'
AVOGADRO = constants.N_A


def polar_partition(value, front_back):
    """ asymmetric partitioning of inclusion body """
    # front aggregate goes to the front daughter
    if 'front' in front_back:
        aggregate = front_back['front']
        return [aggregate, 0.0]
    # back aggregate goes to the back daughter
    elif 'back' in front_back:
        aggregate = front_back['back']
        return [0.0, aggregate]


class InclusionBody(Process):
    ''' Inclusion Body Process

    Simulates the conversion of biomass into cell damage in inclusion bodies.
    The damage builds up in the front and back poles, with a bistable switch
    that makes the damage aggregate in one pole or the other. Upon division,
    a polar_partition _divider pulls the inclusion body on the front pole into
    one daughter and the inclusion body on the back pole to another.
    '''

    name = NAME
    defaults = {
        'aggregation_rate': 1e-1,
        'damage_rate': 1e-6,
        'unit_mw': AVOGADRO * units.fg / units.mol,
    }

    def __init__(self, initial_parameters=None):
        super(InclusionBody, self).__init__(initial_parameters)
        self.aggregation_rate = self.parameters['aggregation_rate']
        self.damage_rate = self.parameters['damage_rate']

    def initial_state(self, config=None):
        if config is None:
            config = {}
        initial_mass = config.get('initial_mass', 0.0)
        front_back = [0.0, initial_mass]
        random.shuffle(front_back)
        state = {
            'front': {'aggregate': front_back[0]},
            'back': {'aggregate': front_back[1]}}
        if 'molecules' in config:
            state['molecules'] = config['molecules']
        return state

    def ports_schema(self):
        return {
            'front': {
                'aggregate': {
                    '_default': 0.0,
                    '_emit': True,
                    '_divider': {
                        'divider': polar_partition,
                        'topology': {'front': ('aggregate',)}},
                    '_properties': {
                        'mw': self.parameters['unit_mw']
                    },
                }
            },
            'back': {
                'aggregate': {
                    '_default': 0.0,
                    '_emit': True,
                    '_divider': {
                        'divider': polar_partition,
                        'topology': {'back': ('aggregate',)}},
                    '_properties': {
                        'mw': self.parameters['unit_mw']
                    },
                }
            },
            'inclusion_body': {
                '_default': 0.0,
                '_emit': True,
                '_updater': 'set',
                '_divider': 'zero',
            },
            'molecules': {
                '*': {
                    '_default': 0.0,
                }
            },
        }

    def next_update(self, timestep, states):

        # get the states
        front_aggregate = states['front']['aggregate']
        back_aggregate = states['back']['aggregate']
        molecules = states['molecules']
        molecule_mass = sum(molecules.values())

        # existing damage aggregate at front or back, pulled towards pole with larger body
        total_aggregate = front_aggregate + back_aggregate
        if total_aggregate > 0:
            combined = front_aggregate * back_aggregate / (total_aggregate ** 2)
            front_aggregation = self.aggregation_rate * \
                                combined * (front_aggregate - back_aggregate) / total_aggregate
            back_aggregation = self.aggregation_rate * \
                               combined * (back_aggregate - front_aggregate) / total_aggregate
        else:
            front_aggregation = total_aggregate
            back_aggregation = total_aggregate

        # proportionate damage to all molecules
        if molecule_mass > 0:
            total_damage = self.damage_rate * molecule_mass * timestep
            delta_molecules = {
                mol_id: - total_damage * mass / molecule_mass
                for mol_id, mass in molecules.items()}
        else:
            total_damage = molecule_mass
            delta_molecules = {}

        # get total change to front and back aggregates
        polar_damage = total_damage / 2
        delta_front = (front_aggregation + polar_damage) * timestep
        delta_back = (back_aggregation + polar_damage) * timestep

        return {
            'front': {
                'aggregate': delta_front},
            'back': {
                'aggregate': delta_back},
            'inclusion_body': (total_aggregate + total_damage),
            'molecules': delta_molecules,
        }


# functions to configure and run the process
def run_inclusion_body(out_dir='out'):

    # initialize the process by passing initial_parameters
    initial_parameters = {'growth_rate': 1e-1}
    inclusion_body_process = InclusionBody(initial_parameters)

    # get initial state
    initial_state = inclusion_body_process.initial_state({
        'initial_mass': 1.0,
        'molecules': {
            'biomass': 1.0}})

    # run the simulation
    sim_settings = {
        'initial_state': initial_state,
        'total_time': 100}
    output = simulate_process_in_experiment(inclusion_body_process, sim_settings)

    # plot the simulation output
    plot_settings = {}
    plot_simulation_output(output, plot_settings, out_dir)


# run module is run as the main program with python vivarium/process/template_process.py
if __name__ == '__main__':
    # make an output directory to save plots
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    run_inclusion_body(out_dir)



from vivarium.library.units import units

toy_chromosome_config = {

    'sequence': 'ATACGGCACGTGACCGTCAACTTA',

    'genes': {
        'oA': ['eA'],
        'oAZ': ['eA', 'eZ'],
        'oB': ['eB'],
        'oBY': ['eB', 'eY']},

    'promoter_order': ['pA', 'pB'],

    'promoters': {
        'pA': {
            'id': 'pA',
            'position': 3,
            'direction': 1,
            'sites': [{
                'position': 0,
                'length': 3,
                'thresholds': {'tfA': 0.3 * units.mM}}],
            'terminators': [
                {
                    'position': 6,
                    'strength': 0.5,
                    'products': ['oA']},
                {
                    'position': 12,
                    'strength': 1.0,
                    'products': ['oAZ']}]},
        'pB': {
            'id': 'pB',
            'position': -3,
            'direction': -1,
            'sites': [{
                'position': 0,
                'length': 3,
                'thresholds': {'tfB': 0.5 * units.mM}}],
            'terminators': [
                {
                    'position': -9,
                    'strength': 0.5,
                    'products': ['oB']},
                {
                    'position': -12,
                    'strength': 1.0,
                    'products': ['oBY']}]}},

    'promoter_affinities': {
        ('pA', None): 1.0,
        ('pA', 'tfA'): 10.0,
        ('pB', None): 1.0,
        ('pB', 'tfB'): 10.0},

    'domains': {
        0: {
            'id': 0,
            'lead': 0,
            'lag': 0,
            'children': []}},

    'rnaps': {}}
